import os
import sys
import subprocess
import shlex  # contains shlex.split that respects quoted strings

# deniz: it is better to use more pathlib in the future so that dir/path
# operations will be more portable (supported since Python 3.4, 2014)
import pathlib

from .software_package import software_package
from esm_parser import user_error

import esm_environment
import esm_plugin_manager


# Yes, Type Hints. Python >= 3.5 supports them. Small steps towards stability,
# until Paul goes crazy and redoes everything in Go. Or Rust. Or Brainfuck
# (yes, that's not made up: https://en.wikipedia.org/wiki/Brainfuck)
# Docs for typing: https://docs.python.org/3/library/typing.html
def install(package: str) -> None:
    """
    Checks if a package is already installed in the system and if it's not, then it
    installs it.

    Parameters
    ----------
    package : str
        Name of the package or get operation. Can be a package name (e.g.
        ``numpy``) or a full pip address (e.g.
        ``git@https://github.com/esm-tools/esm_tools.git``)

    Returns
    -------
    None
    """
    package_name = package.split("/")[-1].replace(".git", "")
    installed_packages = esm_plugin_manager.find_installed_plugins()
    if not package_name in installed_packages:
        try:
            subprocess.check_call([sys.executable, "-m", "pip", "install", "--user",package])
        except (OSError, subprocess.CalledProcessError):  # PermissionDeniedError would be nicer...
            subprocess.check_call(
                [sys.executable, "-m", "pip", "install", "--user", package]
            )


######################################################################################
################################# class "task" #######################################
######################################################################################


class Task:
    """What you can do with a software_package, e.g. comp-awicm-2.0"""

    def __init__(self, raw, setup_info, vcs, general, complete_config, parsed_args):
        if raw == "default":
            raw = ""
        if raw == "drytestall":
            # panic
            for package in setup_info.all_packages:
                for todo in package.targets:
                    try:
                        print(todo + "-" + package.raw_name)
                        newtask = Task(
                            todo + "-" + package.raw_name, setup_info, vcs, parsed_args
                        )
                        newtask.output_steps()
                    except:
                        print("Problem found with target " + newtask.raw_name)
                        sys.exit(1)
            sys.exit(0)

        self.required_plugins = {}
        for key in complete_config:
            if "required_plugins" in complete_config[key]:
                self.required_plugins[key] = complete_config[key]["required_plugins"]
        # Initialize the ``already_installed_plugins`` attribute if it does not exist
        general.already_installed_plugins = self.already_installed_plugins = getattr(
            general, "already_installed_plugins", []
        )

        if type(raw) == str:
            (
                self.todo,
                kind,
                model,
                version,
                self.only_subtask,
                self.raw_name,
            ) = setup_info.split_raw_target(raw, setup_info)
            self.package = software_package(
                (kind, model, version), setup_info, vcs, general
            )
        else:  # tupel:
            (self.todo, kind, model, version, self.only_subtask) = raw
            self.package = software_package(
                (kind, model, version), setup_info, vcs, general
            )
            self.raw_name = setup_info.assemble_raw_name(
                self.todo, kind, model, version
            )

        if kind == "components":
            self.env = esm_environment.esm_environment.EnvironmentInfos(
                "compiletime", complete_config, model
            )
        else:
            self.env = None
        if not self.todo in setup_info.meta_todos:
            self.check_if_target(setup_info)

        self.subtasks = self.get_subtasks(
            setup_info, vcs, general, complete_config, parsed_args
        )
        self.only_subtask = self.validate_only_subtask()
        self.ordered_tasks = self.order_subtasks(setup_info, vcs, general)

        self.will_download = self.check_if_download_task(setup_info)
        self.folders_after_download = self.download_folders()
        self.binaries_after_compile = self.compile_binaries()
        self.dir_list = self.list_required_dirs()
        self.command_list, self.shown_command_list = self.assemble_command_list()

        if parsed_args.get("verbose", False):
            self.output()

    def get_subtasks(self, setup_info, vcs, general, complete_config, parsed_args):
        subtasks = []
        if self.todo in setup_info.meta_todos:
            todos = setup_info.meta_command_order[self.todo]
        else:
            todos = [self.todo]
        for todo in todos:
            for subpackage in self.package.subpackages:
                if todo in subpackage.targets:
                    subtasks.append(
                        Task(
                            (
                                todo,
                                subpackage.kind,
                                subpackage.model,
                                subpackage.version,
                                None,
                            ),
                            setup_info,
                            vcs,
                            general,
                            complete_config,
                            parsed_args,
                        )
                    )
        if subtasks == [] and self.todo in setup_info.meta_todos:
            # if self.todo in setup_info.meta_todos:
            for todo in todos:
                if todo in self.package.targets:
                    subtasks.append(
                        Task(
                            (
                                todo,
                                self.package.kind,
                                self.package.model,
                                self.package.version,
                                None,
                            ),
                            setup_info,
                            vcs,
                            general,
                            complete_config,
                            parsed_args,
                        )
                    )
        return subtasks

    def validate_only_subtask(self):
        only = None
        if self.only_subtask:
            only = []
            for task in self.subtasks:
                if task.package.raw_name.startswith(self.only_subtask):
                    only.append(task)
            if self.package.raw_name.startswith(self.only_subtask):
                self.subtasks = []
                return None
            if only == []:
                print()
                print(
                    "Given subtask "
                    + self.only_subtask
                    + " is not a valid subtask of package "
                    + self.raw_name
                    + "."
                )
                print()
                sys.exit(0)

        return only

    def order_subtasks(self, setup_info, vcs, general):
        subtasks = self.subtasks
        if self.only_subtask:
            if self.only_subtask == "NONE":
                return []
            elif type(self.only_subtask) == str:
                return [self.only_subtask]
            else:
                subtasks = self.only_subtask
        if subtasks == []:
            return [self]
        if self.todo in setup_info.meta_todos:
            todos = setup_info.meta_command_order[self.todo]
        else:
            todos = [self.todo]

        ordered_tasks = []
        for todo in todos:
            for task in subtasks:
                if task.todo == todo and task.package.bin_type == "lib":
                    ordered_tasks.append(task)
            for task in subtasks:
                if task.todo == todo and not task.package.bin_type == "lib":
                    ordered_tasks.append(task)  #
        if self.package.kind == "components" and not self.only_subtask:
            ordered_tasks.append(self)
        return ordered_tasks

    def check_if_download_task(self, setup_info):
        if self.todo == "get":
            return True
        if self.todo in setup_info.meta_todos:
            if "get" in setup_info.meta_command_order[self.todo]:
                return True
        return False

    def download_folders(self):
        # if self.package.kind in ["setups", "couplings"]:
        if self.package.subpackages:
            dir_list = [self.package.raw_name]
            for task in self.ordered_tasks:
                if (
                    self.package.raw_name + "/" + task.package.destination
                    not in dir_list
                ):
                    dir_list.append(
                        self.package.raw_name + "/" + task.package.destination
                    )
        else:
            dir_list = []
            for task in self.ordered_tasks:
                if task.package.destination not in dir_list:
                    dir_list.append(task.package.destination)
        return dir_list

    def compile_binaries(self):
        file_list = []
        for task in self.ordered_tasks:
            for binfile in task.package.bin_names:
                if (
                    self.package.raw_name
                    + "/"
                    + task.package.bin_type
                    + "/"
                    + binfile.split("/")[-1]
                    not in file_list
                ):
                    file_list.append(
                        self.package.raw_name
                        + "/"
                        + task.package.bin_type
                        + "/"
                        + binfile.split("/")[-1]
                    )
        return file_list

    def list_required_dirs(self):
        toplevel = self.package.raw_name
        if self.package.kind in ["setups", "couplings"] and self.will_download:
            dir_list = [self.package.raw_name]
        else:
            dir_list = []
        for task in self.ordered_tasks:
            if task.todo == "comp":
                if task.package.bin_names:
                    newdir = toplevel + "/" + task.package.bin_type
                    if newdir not in dir_list:
                        dir_list.append(newdir)
        return dir_list

    def assemble_command_list(self):
        command_list = []
        toplevel = self.package.destination
        # if self.package.kind in ["setups", "couplings"]:
        if self.package.subpackages:  # ???
            command_list.append("mkdir -p " + toplevel)
            command_list.append("cd " + toplevel)
            toplevel = "."
        real_command_list = command_list.copy()
        for task in self.ordered_tasks:
            if task.todo in ["get"]:
                if task.package.command_list[task.todo] is not None:
                    for command in task.package.command_list[task.todo]:
                        command_list.append(command)
                        real_command_list.append(command)

        if self.package.coupling_changes:
            for change in self.package.coupling_changes:
                command_list.append(change)
                real_command_list.append(change)

        for task in self.ordered_tasks:
            if task.todo not in ["get"]:
                if task.todo in ["conf", "comp"]:
                    # if self.package.kind in ["setups", "couplings"]:
                    if task.package.kind not in ["setups", "couplings"]:
                        if self.package.subpackages:
                            real_command_list.append(
                                "cp ../" + task.raw_name + "_script.sh ."
                            )
                        real_command_list.append("./" + task.raw_name + "_script.sh")
                else:
                    if task.package.command_list[task.todo] is not None:
                        for command in task.package.command_list[task.todo]:
                            real_command_list.append(command)
                if task.package.command_list[task.todo] is not None:
                    for command in task.package.command_list[task.todo]:
                        command_list.append(command)
                if task.todo == "comp":
                    if task.package.bin_names:
                        command_list.append(
                            "mkdir -p " + toplevel + "/" + task.package.bin_type
                        )
                        real_command_list.append(
                            "mkdir -p " + toplevel + "/" + task.package.bin_type
                        )
                        for binfile in task.package.bin_names:
                            # PG: Only copy if source and dest aren't the same!
                            # (Prevents cp: ‘/temp/test.txt’ and
                            # ‘/temp/test/test.txt’ are the same file)
                            toplevel_bin_path = (
                                toplevel
                                + "/"
                                + task.package.bin_type
                                + "/"
                                + binfile.split("/")[-1]
                            )
                            # MA: If there are already commands that will clean the
                            # bin folder, the following `if` is required to be true
                            clean_command_list = ["rm -f " + toplevel_bin_path]
                            clean_command = any(
                                cc in command_list for cc in clean_command_list
                            )

                            # deniz: bug fix for the fesom compilation issue
                            # at some point these string concats need to be
                            # replaced by pathlib
                            bin_path_target = pathlib.Path(
                                toplevel + "/" + task.package.bin_type
                            )
                            binary_file_path = pathlib.Path(
                                task.package.destination + "/" + binfile
                            )
                            binary_file_parent = binary_file_path.parent
                            # path2bin_origin = "/".join(binary_file_path.split('/')[:-1])

                            # deniz: don't copy the files if the paths are same.
                            # I think pathlib.Path.resolve() is a better option
                            # than simple string comparison
                            should_copy_files = (
                                binary_file_parent.resolve()
                                != bin_path_target.resolve()
                            )

                            # add the remaining conditions
                            should_copy_files = should_copy_files and (
                                not os.path.exists(toplevel_bin_path) or clean_command
                            )

                            if should_copy_files:
                                # deniz: I think " ".join() is a more Pythonic
                                # way to construct the command strings
                                cmd_str = "  ".join(
                                    ["cp", str(binary_file_path), str(bin_path_target)]
                                )
                                command_list.append(cmd_str)
                                real_command_list.append(cmd_str)

                elif task.todo == "clean":
                    if task.package.bin_names:
                        for binfile in task.package.bin_names:
                            command_list.append(
                                f"rm -f {toplevel}/{task.package.bin_type}/{binfile.split('/', -1)[-1]}"
                            )
                            real_command_list.append(
                                f"rm -f {toplevel}/{task.package.bin_type}/{binfile.split('/', -1)[-1]}"
                            )
        if task.todo in ["comp"]:
            for component in self.required_plugins:
                for plugin in self.required_plugins[component]:
                    # Install the plugin if is not already installed
                    if plugin not in self.already_installed_plugins:
                        # Actually only works because Paul put the gfw_creator
                        # required plugin for awiesm onto PyPI...
                        install(plugin)
                        self.already_installed_plugins.append(plugin)

        if self.package.kind in ["setups", "couplings"]:
            command_list.append("cd ..")
            real_command_list.append("cd ..")

        return real_command_list, command_list

    def cleanup_script(self):
        try:
            os.remove("./dummy_script.sh")
        except OSError:
            print("No dummy script to remove!")
        for task in self.ordered_tasks:
            if task.todo in ["conf", "comp"]:
                try:
                    os.remove("./" + task.raw_name + "_script.sh")
                except OSError:
                    print("No file to remove for ", task.raw_name)

    def check_if_target(self, setup_info):
        if not setup_info.has_target2(self.package, self.todo):
            setup_info.output_available_targets(self.raw_name)
            sys.exit(0)

    def check_requirements(self):
        if self.will_download:
            return True
        requirements = self.folders_after_download
        for folder in requirements:
            if not os.path.isdir(folder):
                print()
                print(
                    "Missing folder "
                    + folder
                    + " detected. Please run 'make get-"
                    + self.package.raw_name
                    + "' first."
                )
                print()
                sys.exit(0)
        return True

    def validate(self):
        self.check_requirements()

    def execute(self, ignore_errors=False):
        for task in self.ordered_tasks:
            if task.todo in ["conf", "comp"]:
                if task.package.kind == "components":
                    task.env.write_dummy_script()
                    newfile = task.env.add_commands(
                        task.package.command_list[task.todo], task.raw_name
                    )
                    if os.path.isfile(newfile):
                        os.chmod(newfile, 0o755)

        # Calculate the number of get commands for this esm_master operation
        self.num_of_get_commands()
        # Loop through the commands
        for command in self.command_list:
            repo = self.get_repo_properties_from_command(command)
            if command.startswith("mkdir") and ";" not in command:
                # os.system(command)
                subprocess.run(command.split(), check=not ignore_errors)
            elif command.startswith("cp ") and ";" not in command:
                subprocess.run(command.split(), check=not ignore_errors)
            elif command.startswith("cd ") and ";" not in command:
                os.chdir(command.replace("cd ", ""))
            # deniz: add pipe support
            elif "|" in command:
                # if there is a pipe in the command, then separate these in to
                # two parts. Eg. curl foo.tar.gz | tar zx
                curl_command, pipe_command = command.split("|")
                curl_process = subprocess.Popen(
                    curl_command.split(), stdout=subprocess.PIPE
                )
                output = subprocess.check_output(
                    pipe_command.split(), stdin=curl_process.stdout
                )
                curl_process.wait()
            elif repo["is_repo_operation"]:
                command_spl = shlex.split(command)
                try:
                    subprocess.run(command_spl, check=not ignore_errors)
                except subprocess.CalledProcessError as error:
                    self.add_repo_error(command, repo, error)
                # If it's the last repo command, check for errors and report them back
                if repo["is_last_repo"]:
                    self.report_repo_errors()
                    self.report_destination_path_errors()
            else:
                # os.system(command)
                # deniz: I personally did not like the iterator and the list
                # having the same name. for com in command.split(';') would be
                # better IMHO
                for command in command.split(";"):
                    # seb-wahl: use shlex split as sed commands that use spaces
                    # need to be quoted, shlex split doesn't split quoted
                    # strings on spaces
                    # example: sed -i '/COUPLENEMOFOCI = /s/.FALSE./.TRUE./g' oifs-43r3-foci/src/ifs/module/yommcc.F90
                    # will fail if the "'" is removed
                    command_spl = shlex.split(command)
                    if "cd" == command_spl[0]:
                        os.chdir(command_spl[1])
                    else:
                        subprocess.run(
                            command_spl,
                            check=True,
                            shell=(
                                command.startswith("./") and command.endswith(".sh")
                            ),
                        )

    def output(self):
        print()
        subtasklist = []
        osubtasklist = []
        self.package.output()
        print("    Todo: ", self.todo)
        if self.only_subtask:
            if self.only_subtask == "NONE":
                print("    NO VALID SUBTASKS!!!")
            else:
                for subtask in self.only_subtask:
                    print("    Only Subtask:", subtask.package.raw_name)
        for subtask in self.subtasks:
            subtasklist.append(subtask.raw_name)
        for osubtask in self.ordered_tasks:
            osubtasklist.append(osubtask.raw_name)
        if not subtasklist == []:
            print("    Subtasks:", subtasklist)
        if not osubtasklist == []:
            print("    Ordered Subtasks:", osubtasklist)
        self.output_steps()
        if not self.folders_after_download == []:
            print("    The following folders should exist after download:")
            for folder in self.folders_after_download:
                print("        ", folder)
        if not self.binaries_after_compile == []:
            print("    The following files should exist after compiling:")
            for binfile in self.binaries_after_compile:
                print("        ", binfile)

    def output_steps(self):
        if not self.command_list == []:
            print("    Executing commands in this order:")
            for command in self.shown_command_list:
                print("        ", command)

    # Repository methods
    # ------------------
    def num_of_get_commands(self):
        """
        Defines ``self.num_get_commands`` which accounts for the total number of get
        commands for the current ``esm_master`` operation
        """
        self.num_get_commands = 0
        for command in self.command_list:
            for subtask in self.subtasks:
                if command in subtask.package.command_list.get("get", []):
                    self.num_get_commands += 1
                    break
        return self.num_get_commands # (Not strictly needed, but might be nice?)

    def get_repo_properties_from_command(self, command):
        """
        If the current command is a repo action (e.g. cloning) collects the repo
        information associated to that command

        Parameters
        ----------
        command : str
            The shell command to be evaluated

        Returns
        -------
        repo : dict
            A dictionary containing the following information:
            - ``package``: the package object that contains all the package info
            - ``is_repo_operation``: boolean that indicates whether ``command`` is a
              repo operation
            - ``is_last_repo``: boolean that indicates whether this is the last repo
              operation or not
        """
        # Initializes the list of executed repo commands
        if not hasattr(self, "executed_repo_commands"):
            self.executed_repo_commands = []

        repo = {"is_repo_operation": False, "is_last_repo": False}
        get_commands = self.package.command_list.get("get")
        # This is true for standalone models
        if self.package.repo and get_commands:
            if command in get_commands:
                repo["package"] = self.package
                repo["is_repo_operation"] = True
                repo["is_last_repo"] = True
        # This block here is for coupled setups (loops through each subtask -> model)
        else:
            for subtask in self.subtasks:
                get_commands = subtask.package.command_list.get("get")
                if command in get_commands:
                    repo ["package"]= subtask.package
                    repo["is_repo_operation"] = True
                    self.executed_repo_commands.append(command)
                    if self.num_get_commands == len(self.executed_repo_commands):
                        repo["is_last_repo"] = True
                    break

        return repo

    def add_repo_error(self, command, repo, error):
        """
        If an error occurred during the repo ``command``, then stores information about
        the error, and the associated package. The information about the error is
        saved in the attribute ``self.repo_errors`` (if it doesn't exist, it creates
        it).

        The ``repo_errors`` attribute itself is a ``dict`` with keys being the failed
        ``command``s, each of them containing the following variables:
        - ``model``: model name
        - ``repo``: url of the repo
        - ``destination``: destination folder
        - ``error``: "destination path exist" if the destination exist, and the value
          of the ``error`` parameter if it the destination does not exist

        Note
        ---- 
        Since the ``error`` is caught from an ``except`` of a ``subprocess``
        command, the actual error that occurred during the execution of the ``command``
        by ``subprocess.run`` is not caught. That means that we cannot evaluate which
        type of error occurred during the repo command (i.e. was an error related to
        access to the repo, or was it the error that happens when the folder already
        exist?). To compensate for this problem, this method also checks whether the
        destination folder exist or not to save a ``"destination path exist"`` error
        description or to keep the same error string as reported by ``error``
        (permission problem). This approach is not ideal. It would be possible to catch
        the ``subprocess`` error, by capturing the stdout/stderr of the ``subprocess``
        command, but then, the downloading progress is also captured and not displayed
        to the user, which might make the users wonder if ESM-Tools is stalling.

        Parameters
        ----------
        command : str
            Repo command for which the error might have occur
        repo : dict
            Information associated to the repo ``command``, including tha ``package``
            object
        error : str
            Error caught from ``except``
        """
        # Initializes the list of executed repo commands
        if not hasattr(self, "repo_errors"):
            self.repo_errors = {}

        destination = repo["package"].destination
        # Checks whether the destination path exists to save a "destination path exist"
        # error
        if os.path.isdir(destination):
            error = "destination path exists"
        for folder in self.folders_after_download:
            if destination in folder:
                full_destination = folder

        # If an error occured, store the information in the ``repo_errors`` atribute
        if error:
            self.repo_errors[command] = {
                "model": repo["package"].model,
                "repo": repo["package"].repo,
                "contact": repo["package"].contact,
                "destination": full_destination,
                "error": error,
            }

    def report_repo_errors(self):
        """
        Reports an ``esm_parser.user_error`` summarizing all the information about
        the repo permission errors associated to the ``esm_master`` command.
        """
        if not hasattr(self, "repo_errors") or len(self.repo_errors) == 0:
            return

        problematic_repos = ""
        for command, repo_error in self.repo_errors.items():
            error = repo_error["error"]
            if error != "destination path exists":
                problematic_repos += (
                    f"``{repo_error['model']}``: {repo_error['repo']}\n"
                    f"    ``contact``: {repo_error['contact']}\n"
                    f"    failed command: ``{command}``\n"
                    f"    error message:\n{error}\n"
                )

        if problematic_repos:
            user_error(
                "Download error",
                "There were problems downloading some of the components involved in "
                f"the operation ``{self.raw_name}``. Make sure the ``user names`` and "
                "``passwords`` you introduced are correct. Also, check that you have "
                "access and reading permissions to the repositories listed below. If "
                "you don't, contact the person in charge of that particular repository "
                "(see ``contact`` in the repository list below).\n\n"
                f"Repositories with problems:\n{problematic_repos}"
            )

    def report_destination_path_errors(self):
        """
        Reports an ``esm_parser.user_error`` summarizing all the information about
        the ``"destination path exist"`` errors associated to the ``esm_master``
        command.
        """
        if not hasattr(self, "repo_errors") or len(self.repo_errors) == 0:
            return

        problematic_destinations = ""
        for command, repo_error in self.repo_errors.items():
            error = repo_error["error"]
            if error == "destination path exists":
                problematic_destinations += f"- ``{repo_error['destination']}``\n"

        if problematic_destinations:
            user_error(
                "destination already exists",
                "Some coponents couldn't be downloaded because their destination paths "
                f"already exists (see list below). If you want to download again the "
                "model consider deleting that folder. If instead, you want to keep "
                "that, you can use other esm_master commands (e.g. esm_master "
                "comp-<model>-<version>). Destinations already present:\n"
                f"{problematic_destinations}\n"
            )



"""
Functionality for displaying the version number
"""
import sys
import click
import esm_tools
import esm_tests

# click.version_option read the PKG_INFO which contains the wrong version
# number. Get it directly from __init__.py
version = esm_tools.__version__

@click.version_option(version=version)
@click.option("--test-state", is_flag=True, help="Prints the state of the last tested experiments.")
@click.command()
def main(test_state):
    """Console script for esm_tools"""

    if test_state:
        esm_tests.test_utilities.print_state_online()

    return 0


if __name__ == "__main__":
    sys.exit(main())



import sys
from datetime import datetime

import colorama
import esm_parser
import esm_plugin_manager
import esm_tools
import git
from loguru import logger


def vprint(message, config):
    if config["general"]["verbose"]:
        print(message)


# TODO: to be replaced by loguru. WIP (deniz)
def print_datetime(config):
    """prints the datetime of the operation if `verbose_line_numbers` option is True"""
    if config["general"].get("verbose_datetime_info", False):
        print(datetime.now(), flush=True)


def evaluate(config, job_type, recipe_name):

    # Check for a user defined compute recipe in the setup section of the
    # general section. If nothing is found, recipe_steps should evaluate to
    # None and the default is used
    try:
        setup_name = config["general"]["setup_name"]
        recipe_steps = config.get(setup_name, {}).get(recipe_name) or config[
            "general"
        ].get(recipe_name)
    except KeyError:
        print("Your configuration is incorrect.")
        print("It should include headings for the setup_name as well as general!")
        sys.exit(1)

    FUNCTION_PATH = esm_tools.get_config_filepath()

    ####################################################################################
    # LA: hotfix for non-matching python paths
    if (
        "/global/AWIsoft/tkleiner/lib/python2.7/site-packages/netCDF4-1.2.4-py2.7-linux-x86_64.egg"  # noqa: E501
        in sys.path
    ):
        sys.path.remove(
            "/global/AWIsoft/tkleiner/lib/python2.7/site-packages/netCDF4-1.2.4-py2.7-linux-x86_64.egg"  # noqa:501
        )
    if "/global/AWIsoft/tkleiner/lib/python2.7/site-packages" in sys.path:
        sys.path.remove("/global/AWIsoft/tkleiner/lib/python2.7/site-packages")
    ####################################################################################

    recipe = FUNCTION_PATH + "/esm_software/esm_runscripts/esm_runscripts.yaml"
    need_to_parse_recipe = True
    plugins_bare = FUNCTION_PATH + "/esm_software/esm_runscripts/esm_plugins.yaml"
    need_to_parse_plugins = True

    framework_recipe = esm_plugin_manager.read_recipe(
        recipe, {"job_type": job_type}, need_to_parse_recipe
    )
    if recipe_steps:
        framework_recipe["recipe"] = recipe_steps
    framework_plugins = esm_plugin_manager.read_plugin_information(
        plugins_bare, framework_recipe, need_to_parse_plugins
    )
    esm_plugin_manager.check_plugin_availability(framework_plugins)

    config = esm_plugin_manager.work_through_recipe(
        framework_recipe, framework_plugins, config
    )
    return config


########################################################################################
#                                  general stuff                                       #
########################################################################################
def end_it_all(config):
    if config["general"]["profile"]:
        for line in timing_info:
            print(line)
    if config["general"]["verbose"]:
        print("Exiting entire Python process!")
    sys.exit()


def write_to_log(config, message, message_sep=None):
    """
    Puts a message into the experiment log file

    Parameters
    ----------
    message : list
        A list of the message elements; which is joined by either (highest
        to lowest): 1) the message_sep argument passed to the method, 2)
        The user's chosen separator, as written in
        ``config["general"]["experiment_log_file_message_sep"]``, 3)
        An empty space ``" "``.
    message_sep : None
        The hard-coded message separator to use; which ignores user choices.

    Note
    ----
    The user can control two things regarding the logfile format:

    1) The datestamp formatting, whjich is taken from the config
       section ``general.experiment_log_file_dateformat``.
    2) The message separators; taken from
       ``general.experiment_log_file_message_sep``. Note that if the
       programmer passes a ``message_sep`` argument; this one wins over
       the user choice.
    """
    try:
        with open(config["general"]["experiment_log_file"], "a+") as logfile:
            line = assemble_log_message(config, message, message_sep)
            logfile.write(line + "\n")
    except KeyError:
        import esm_parser

        print("Sorry; couldn't find 'experiment_log_file' in config['general']...")
        esm_parser.pprint_config(config["general"])
        raise


def assemble_log_message(
    config, message, message_sep=None, timestampStr_from_Unix=False
):
    """Assembles message for log file. See doc for write_to_log"""
    message = [str(i) for i in message]
    dateTimeObj = datetime.now()
    strftime_str = config["general"].get("experiment_log_file_dateformat", "%c")
    if message_sep is None:
        message_sep = config["general"].get("experiment_log_file_message_sep", " ")
    if timestampStr_from_Unix:
        if strftime_str == "%c":
            # date doesn't do %c as it is supposed to
            strftime_str = '"%a %b  %e %T %Y"'
        timestampStr = "$(date +" + strftime_str + ")"
    else:
        timestampStr = dateTimeObj.strftime(strftime_str)
    # TODO: Do we want to be able to specify a timestamp separator as well?
    line = timestampStr + " : " + message_sep.join(message)
    return line


def update_reusable_filetypes(config, reusable_filetypes=None):
    """
    Removes some filetypes from the "resuable files" list. Filetypes to be
    removed are either specified in the config, or passed as an argument.

    Reusable filetypes, declared in general
    (``config["general"]["reusable_filetypes"]``), or in model specific
    (``config[<model>]["reusable_filetypes"]``) forces files of a given type,
    such as ``bin``, ``input``, ``forcing``, etc., to be copied only once into
    the general experiment folder, and then **reuse** them through out every
    run. If a given file type is not present in this list, that file type will
    be copied from its external source at the beginning of every run (i.e. for
    binaries from the folder where the model is compiled).

    This method takes care of removing the reusable file types from
    ``reusable_filetypes``, specified by the user in the ``esm_runscripts``
    call with the ``--update-filetypes`` flag.

    Parameters
    ----------
    config : dict
        Dictionary containing the information about the experiment.
    reusable_filetypes : list
        List of reusable file types that can be specified by the user.

    Returns
    -------
    dict or list
        If the user gives no ``reusable_filetypes`` as input. A dict
        ``config`` is returned, which contains the changes requested
        by the user with the ``--update-flag``.

        If the user gives ``reusable_filetypes`` as input, a list is
        returned. This list ``reusable_filetypes`` contains the changes
        requested by the user with the ``--update-flag``.
    """
    # Set the logic for what to return
    if reusable_filetypes:
        return_config = False
    else:
        return_config = True
        reusable_filetypes = config["general"].get("reusable_filetypes", [])

    update_filetypes = (
        config["general"].get("command_line_config", {}).get("update_filetypes", [])
    )
    if config["general"].get("verbose", False) and update_filetypes:
        print("User requests that the following filetypes be updated:")
        [print(f"* {filetype}") for filetype in update_filetypes]

    # NOTE(MAM, PG): Originally defined in prepare.py
    # https://tinyurl.com/2p8awzsu
    potentially_reusable_filetypes = config["general"]["potentially_reusable_filetypes"]

    # Loop through the file types specified by the user with the ``--update-filetypes``
    # flag
    for update_filetype in update_filetypes:
        # Check if that file type exists/makes sense. Otherwise, through an error
        if update_filetype not in potentially_reusable_filetypes:
            esm_parser.user_error(
                "update-filetypes",
                f"``{update_filetype}`` specified by you in ``--update-filetypes`` is "
                + "not a ESM-Tools file type. Please, select one (or more) of the "
                + "following file types:\n\t- "
                + "\n\t- ".join(potentially_reusable_filetypes),
            )
        # Actually remove the file types specified by the user from
        # ``reusable_filetypes``
        if update_filetype in reusable_filetypes:
            # Remove duplicates just in case
            reusable_filetypes = list(set(reusable_filetypes))
            # Do the removal
            if config["general"].get("verbose", False):
                print(f"Removing {update_filetype}")
            reusable_filetypes.remove(update_filetype)
        elif config["general"].get("verbose", False):
            print(
                f"- The file type ``{update_filetype}`` you are trying to update was"
                "not reusable across runs in the first place, so it's been always "
                "updated with the external source, and it will still be."
            )

    if config["general"].get("verbose", False):
        print("The following filetypes will be re-used from already copied sources:")
        for reusable_filetype in reusable_filetypes:
            print(f"* {reusable_filetype}")
    config["general"]["reusable_filetypes"] = reusable_filetypes

    if return_config:
        return config
    return reusable_filetypes


##############################
# SINK CLASS FOR LOGURU.LOGGER
##############################


class SmartSink:
    """
    A class for smart sinks that allow for logging (using ``logger`` from loguru), even
    if the file path of the log file is not yet defined. The actual sink is not the
    instanced object itself, but the method ``sink`` of the instance. The log record is
    saved in ``self.log_record`` and the log file is written using the path specified
    in ``self.path``. If the path is not specified, the log is stored only in the
    ``self.log_record``. When the path is finally specified, ``self.log_record`` is
    dumped into the log file and from that moment, any time ``logger`` logs something it
    will also be written into the file. To specify the path the method ``def_path``
    needs to be used.
    """

    def __init__(self):
        # Initialise instance variables
        self.log_record = []
        self.path = None

    def sink(self, message):
        """
        The actual sink for loguru's ``logger``. Once you define a logger level a sink
        needs to be provided. Standard sinks include file paths, methods, etc.
        Providing this method as a sink (``logger.add(<name_of_the_instance>.sink,
        level="<your_level>", ...)``) enables the functionality of the SmartSink object.

        Parameters
        ----------
        message : str
            String containing the logging message.
        """
        if self.path:
            self.write_log(message, "a")
        self.log_record.append(message)

    def write_log(self, message, wmode):
        """
        Method to write the logs into the disk.

        Parameters
        ----------
        message : str, list
            String containing the logging message or list containing more than one
            logging message, to be written in the file.
        wmode : str
            Writing mode to choose among ``"w"`` or ``"a"``.
        """
        if isinstance(message, str):
            message = [message]
        with open(self.path, wmode) as log:
            for line in message:
                log.write(line)

    def def_path(self, path):
        """
        Method to define the path of the file. Once the path is defined, the log record
        is written into the file.

        Parameters
        ----------
        path : str
            Path of the logging file.
        """
        self.path = path
        self.write_log(self.log_record, "w")


################################################################################
# Git Checks of esm-tools
#
# NOTE(PG): These functions would likely be better as a separate file


def is_git_repo(path):
    """
    Determines whether or not a directory is a git repository.

    Parameters
    ----------
    path : str or pathlib.Path
        The location to check

    Returns
    bool :
        True or False, depending on if the location is a git directory
    """
    try:
        git.Repo(path).git_dir
        return True
    except git.exc.InvalidGitRepositoryError:
        return False
    except git.exc.NoSuchPathError:
        logger.error("You specified a non-existant directory")
        # Is False the right answer here? I am not 100% sure...
        return False


class GitDirtyError(git.exc.GitError):
    """Thrown if the git repository is dirty"""


def get_git_hash(path, allow_dirty=True):
    """
    Gets the commit has of a git directory stored at ``path``

    Parameters
    ----------
    path : str or pathlib.Path
        Path to determine the git hash for
    allow_dirty : bool
        Complain or not complain about a dirty git repo.

    Returns
    -------
    githash : str
        The commit has the repository is currently on

    Raises
    ------
    GitDirtyError :
       Raised if the repo is "dirty"; thus meaning that the hash is not meaningful
    """
    repo = git.Repo(path)
    if not allow_dirty and repo.is_dirty():
        raise GitDirtyError(
            f"Your repo at {path} is dirty, thus the git hash is not meaningful!"
        )
    return repo.git.rev_parse(repo.head, short=True)


def get_git_branch(path):
    """Gets the name of the current git branch for repo at path

    Parameters
    ----------
    path : str
        The path of the repository to examine

    Returns
    -------
    str :
        The branch name
    """
    repo = git.Repo(path)
    if not repo.head.is_detached:
        return repo.head.reference.name
    commit = repo.head.commit
    for tag in repo.tags:
        if commit.hexsha == tag.commit.hexsha:
            return f"DETACHED HEAD at tag {tag.name}"
    else:
        return f"DETACHED HEAD at commit {repo.head.commit.hexsha}"


def get_git_diffs(path, add_colors={"+": colorama.Fore.GREEN, "-": colorama.Fore.RED}):
    """
    Gets the differences

    Parameters
    ----------
    path : str
        The git repository to check
    add_colors : dict
        A dictionary of ``colorama.Fore`` properties for added lines, stored
        under the key ``'+'`` and removed lines, stored under the key ``'-'``.

    Returns
    -------
    str :
        The (maybe) colorized difference as a multi-line string.
    """
    repo = git.Repo(path)
    diffs = repo.git.diff(repo.commit()).split("\n")
    if add_colors:
        for index, diff in enumerate(diffs):
            if diff.startswith("+"):
                diff = f'{add_colors["+"]}{diff}{colorama.Style.RESET_ALL}'
            elif diff.startswith("-"):
                diff = f'{add_colors["-"]}{diff}{colorama.Style.RESET_ALL}'
            diffs[index] = diff
    # In case no diffs are detected, we still want to return an empty
    # list.
    if not diffs:
        diffs = "\n".join([])
    return "\n".join(diffs)


def get_all_git_info(path):
    """Gets all information needed for the vcs dump file

    Returns
    -------
    dict :
        A dictionary with the following information: git hash, branch name, differences, .....
    """
    git_info = {
        "path": path,
        "hash": get_git_hash(path),
        "branch_name": get_git_branch(path),
        # NOTE(PG): Dumping to YAML is not friendly with colors, it seems...
        "diffs": get_git_diffs(path, add_colors=False),
    }
    return git_info



import os, sys, re
import copy
import asyncio

import pickle

from .general_stuff import (
    COMPONENTS_DIR,
    COUPLINGS_DIR,
    SETUPS_DIR,
    DEFAULTS_DIR,
    ESM_SOFTWARE_DIR,
    ESM_MASTER_PICKLE,
    ESM_MASTER_DIR,
)

import esm_parser
from .software_package import *
import colorama

######################################################################################
############################## Combine all YAMLS #####################################
######################################################################################


def save_pickle(obj, path):
    with open(path, "wb") as f:
        pickle.dump(obj, f, pickle.HIGHEST_PROTOCOL)


def load_pickle(path):
    if os.path.isfile(path):
        with open(path, "rb") as f:
            return pickle.load(f)
    else:
        return None


def combine_components_yaml(parsed_args):
    """
    Combines various YAML files in esm_master config directory.

    Returns
    -------
    dict :
        A dictionary equivalent of all components, couplings, setups, and
        general information.
    """

    relevant_entries = [
        "git-repository",
        "curl-repository",  # deniz
        "repo_options",  # kh 11.09.20 support git options like --recursive
        "branch",
        "tag",
        "comp_command",
        "conf_command",
        "clean_command",
        "components",
        "coupling_changes",
        "requires",
        "couplings",
        "install_bins",
        "install_libs",
        "destination",
        "clone_destination",
        "archfile",
        "use_oasis",
        "pipe_options",  # deniz: Linux pipe support
        "contact",
    ]

    categories = ["components", "couplings", "setups", "esm_software"]

    relevant_dirs = {
        "components": COMPONENTS_DIR,
        "couplings": COUPLINGS_DIR,
        "setups": SETUPS_DIR,
        "esm_software": ESM_SOFTWARE_DIR,
    }

    components_dict = {}

    for cat in categories:

        components_dict[cat] = {}
        cat_dir = relevant_dirs[cat]

        # for package in os.listdir(cat_dir):

        asyncio.get_event_loop().run_until_complete(
            get_all_package_info(
                os.listdir(cat_dir),
                cat,
                cat_dir,
                components_dict,
                relevant_entries,
                parsed_args,
            )
        )
        # TODO(PG): Switch around async optional
        # get_all_package_info(os.listdir(cat_dir), cat, cat_dir, components_dict, relevant_entries, parsed_args)
    default_infos = {}

    # If general.yaml file exits ignore it
    yaml_files = os.listdir(DEFAULTS_DIR)
    if "general.yaml" in yaml_files:
        yaml_files.remove("general.yaml")

    for yaml_file in yaml_files:
        if os.getenv("ESM_MASTER_DEBUG"):
            print(f"Reading file {DEFAULTS_DIR}/{yaml_file}")
        file_contents = esm_parser.yaml_file_to_dict(f"{DEFAULTS_DIR}/{yaml_file}")
        default_infos.update(file_contents)

    components_dict["defaults"] = default_infos

    # esm_parser.pprint_config(components_dict)
    # sys.exit(0)
    return components_dict, relevant_entries


async def get_all_package_info(
    packages, cat, cat_dir, components_dict, relevant_entries, parsed_args
):
    # TODO(PG): Switch around async optional
    # def get_all_package_info(packages, cat, cat_dir, components_dict, relevant_entries, parsed_args):
    tasks = []
    # TODO(PG): Better logging (see GH Issue #116)
    if os.getenv("ESM_MASTER_DEBUG"):
        print(f"packages={packages}")
    for package in packages:
        # TODO(PG): Better logging (see GH Issue #116)
        if os.getenv("ESM_MASTER_DEBUG"):
            print(f"Getting {package}")
        # TODO(PG): Switch around async optional
        # task = get_one_package_info(package, cat, cat_dir, components_dict, relevant_entries)
        task = asyncio.ensure_future(
            get_one_package_info(
                package, cat, cat_dir, components_dict, relevant_entries, parsed_args
            )
        )
        tasks.append(task)
    # TODO(PG): Switch around async optional
    # return tasks
    await asyncio.gather(*tasks, return_exceptions=False)


async def get_one_package_info(
    package, cat, cat_dir, components_dict, relevant_entries, parsed_args
):
    # TODO(PG): Switch around async optional
    # def get_one_package_info(package, cat, cat_dir, components_dict, relevant_entries, parsed_args):

    # TODO(PG): Better logging (see GH Issue #116)
    if os.getenv("ESM_MASTER_DEBUG"):
        print(f"Working on package={package}, cat={cat}, cat_dir={cat_dir}")

    package_dir = cat_dir + package + "/"

    default_file = package_dir + package + ".yaml"
    # TODO(PG): Better logging (see GH Issue #116)
    if os.getenv("ESM_MASTER_DEBUG"):
        print(f"default_file={default_file}")

    versioned_files = [
        package_dir + i
        for i in os.listdir(package_dir)
        if i.startswith(package + "-")
        if i.endswith(".yaml")
    ]
    # TODO(PG): Better logging (see GH Issue #116)
    if os.getenv("ESM_MASTER_DEBUG"):
        print(f"versioned_files={versioned_files}")

    comp_config = esm_parser.yaml_file_to_dict(default_file)
    # TODO(PG): Better logging (see GH Issue #116)
    if os.getenv("ESM_MASTER_DEBUG"):
        if not comp_config:
            print(f"Whoops, got False-y thingy!")
    if os.getenv("ESM_MASTER_DEBUG"):
        print(f"...reading file {default_file}")
    if get_correct_entry(comp_config, {}, "version") == {}:
        if os.getenv("ESM_MASTER_DEBUG"):
            print(f'Var "version" is missing in yaml file for package {package}. ')
            print('Trying to set to "*"...')
        comp_config["version"] = "*"

    package_conf = get_relevant_info(relevant_entries, comp_config)

    for conf_file in versioned_files:
        if os.getenv("ESM_MASTER_DEBUG"):
            print(f"...reading file {conf_file}")
        add_config = esm_parser.yaml_file_to_dict(conf_file)
        if get_correct_entry(add_config, {}, "version") == {}:
            if os.getenv("ESM_MASTER_DEBUG"):
                print(f'Var "version" is missing in yaml file for package {package}. ')
                print('Trying to set to "*"...')
            add_config["version"] = "*"
        package_conf = get_relevant_info(relevant_entries, add_config, package_conf)

    package_conf = remove_globbing_char(package_conf)
    if not package_conf == {}:
        components_dict[cat][package] = package_conf


def remove_globbing_char(conf):
    if "available_versions" in conf and conf["available_versions"] == ["*"]:
        if "choose_version" in conf and list(conf["choose_version"].keys()) == ["*"]:
            if not conf["choose_version"]["*"] == {}:
                conf.update(conf["choose_version"]["*"])
            del conf["choose_version"]
            del conf["available_versions"]

    if "choose_version" in conf and "*" in conf["choose_version"]:
        if "available_versions" in conf:
            for entry in conf["available_versions"]:
                if not entry in conf["choose_version"]:
                    conf["choose_version"].update({entry: conf["choose_version"]["*"]})
            del conf["choose_version"]["*"]

    if "available_versions" in conf:
        if "choose_version" in conf:
            for entry in conf["choose_version"]:
                if not entry in conf["available_versions"]:
                    conf["available_versions"].append(entry)
        if "*" in conf["available_versions"]:
            conf["available_versions"].remove("*")

    if "choose_version" in conf:
        for entry in list(conf["choose_version"].keys()):
            if conf["choose_version"][entry] == {}:
                del conf["choose_version"][entry]
                if "available_versions" in conf:
                    if entry in conf["available_versions"]:
                        conf["available_versions"].remove(entry)
        if conf["choose_version"] == {}:
            del conf["choose_version"]

    if "available_versions" in conf and conf["available_versions"] == []:
        del conf["available_versions"]

    return conf


def get_correct_entry(in_config, out_config, entry, default=None):
    compile_tag = "compile_infos"

    if compile_tag in in_config and entry in in_config[compile_tag]:
        out_config[entry] = in_config[compile_tag][entry]
    elif (
        "general" in in_config
        and compile_tag in in_config["general"]
        and entry in in_config["general"][compile_tag]
    ):
        out_config[entry] = in_config["general"][compile_tag][entry]
    elif "general" in in_config and entry in in_config["general"]:
        out_config[entry] = in_config["general"][entry]
    elif entry in in_config:
        out_config[entry] = in_config[entry]
    else:
        if default:
            out_config[entry] = default

    return out_config


def get_relevant_info(relevant_entries, raw_config, merge_into_this_config=None):
    """
    Gets relevant information from the raw configuration and update the given
    configuration dictionary ``merge_into_this_config``.

    Parameters
    ----------
    relevant_entries : list
        A list of relevant entries from which information needs to be extracted.
    raw_config : dict
        A dictionary containing the raw information read from the `yaml` file.
    merge_into_this_config : dict
        A dictionary in which the relevant information will be added.

    Returns
    -------
    merge_into_this_config : dict
        A dictionary given as input, then updated with the relevant information.
    """

    relevant_info = {}
    for entry in relevant_entries:
        relevant_info = get_correct_entry(raw_config, relevant_info, entry)

    # Load default version from the raw configuration and turn it into a string
    default_version = get_correct_entry(raw_config, {}, "version")["version"]
    default_version = str(default_version)

    comp_config = get_correct_entry(
        raw_config, {}, "available_versions", [default_version]
    )
    comp_config = get_correct_entry(
        raw_config, comp_config, "choose_version", {default_version: {}}
    )

    if default_version not in comp_config["choose_version"]:
        comp_config["choose_version"][default_version] = {}

    for version in comp_config["choose_version"]:
        for entry, value in relevant_info.items():
            if not entry in comp_config["choose_version"][version]:
                comp_config["choose_version"][version][entry] = value
        for entry in list(comp_config["choose_version"][version].keys()):
            if entry not in relevant_entries:
                del comp_config["choose_version"][version][entry]

    if merge_into_this_config:
        for version in comp_config["available_versions"]:
            if version not in merge_into_this_config["available_versions"]:
                merge_into_this_config["available_versions"].append(version)

        for version in comp_config["choose_version"]:
            if version in merge_into_this_config["choose_version"]:
                print(f"Error: Version {version} defined two times.")
                sys.exit(-1)
        merge_into_this_config["choose_version"].update(comp_config["choose_version"])

    else:
        merge_into_this_config = copy.deepcopy(comp_config)

    return merge_into_this_config


######################################################################################
########################### class "setup_and_model_infos" ############################
######################################################################################


class setup_and_model_infos:
    def __init__(self, vcs, general, parsed_args):

        if not os.path.isfile(ESM_MASTER_PICKLE):
            self.config, self.relevant_entries = combine_components_yaml(parsed_args)
            save_pickle(self.config, ESM_MASTER_PICKLE)

        elif "list_all_packages" in parsed_args:
            self.config = load_pickle(ESM_MASTER_PICKLE)

        else:
            self.config, self.relevant_entries = combine_components_yaml(parsed_args)
            save_pickle(self.config, ESM_MASTER_PICKLE)

        self.model_kinds = list(self.config.keys())
        self.meta_todos = general.meta_todos
        self.meta_command_order = general.meta_command_order
        self.display_kinds = general.display_kinds

        self.model_todos = []
        for kind in self.model_kinds:
            for model in self.config[kind].keys():
                version = None
                if "choose_version" in self.config[kind][model]:
                    for version in self.config[kind][model]["choose_version"]:
                        for entry in self.config[kind][model]["choose_version"][
                            version
                        ]:
                            if entry.endswith("_command"):
                                todo = entry.replace("_command", "")
                                if todo not in self.model_todos:
                                    self.model_todos.append(todo)
                for entry in self.config[kind][model]:
                    if entry.endswith("_command"):
                        todo = entry.replace("_command", "")
                        if todo not in self.model_todos:
                            self.model_todos.append(todo)

        self.known_todos = self.model_todos + vcs.known_todos + general.meta_todos

        self.all_packages = self.list_all_packages(vcs, general)
        self.update_packages(vcs, general)

        if parsed_args.get("verbose", False):
            self.output()

    def append_to_conf(self, target, reduced_config, toplevel=""):
        (todo, kind, model, version, only_subtarget, raw) = self.split_raw_target(
            target, self
        )
        if not version:
            version = "default"

        if model in self.config[kind]:
            reduced_config[model] = self.config[kind][model]
            reduced_config[model]["version"] = version
            reduced_config[model]["kind"] = kind
            esm_parser.choose_blocks(reduced_config)
        if kind == "setups":
            toplevel = model + "-" + version
            reduced_config[model]["model_dir"] = ESM_MASTER_DIR + "/" + toplevel
            if "couplings" in self.config[kind][model]:
                for coupling in self.config[kind][model]["couplings"]:
                    reduced_config = self.append_to_conf(
                        coupling, reduced_config, toplevel
                    )
        elif kind == "couplings":
            if toplevel == "":
                toplevel = model + "-" + version
            reduced_config[model]["model_dir"] = ESM_MASTER_DIR + "/" + toplevel
            if "components" in self.config[kind][model]:
                for component in self.config[kind][model]["components"]:
                    reduced_config = self.append_to_conf(
                        component, reduced_config, toplevel
                    )
        elif kind == "components":
            sep = ""
            if toplevel == "":
                if "requires" in self.config[kind][model]:
                    toplevel = model + "-" + version
                    sep = "/"
            else:
                sep = "/"

            if "destination" in reduced_config[model]:
                reduced_config[model]["model_dir"] = (
                    ESM_MASTER_DIR
                    + "/"
                    + toplevel
                    + sep
                    + reduced_config[model]["destination"]
                )
            else:
                reduced_config[model]["model_dir"] = (
                    ESM_MASTER_DIR + "/" + toplevel + sep + model + "-" + version
                )

            if "requires" in self.config[kind][model]:
                for requirement in self.config[kind][model]["requires"]:
                    reduced_config = self.append_to_conf(
                        requirement, reduced_config, toplevel
                    )

        return reduced_config

    # def reduce(self, target, env):
    def reduce(self, target):
        blacklist = [re.compile(entry) for entry in ["computer.*"]]

        reduced_config = {}
        reduced_config["defaults"] = self.config["defaults"]
        reduced_config = self.append_to_conf(target, reduced_config)

        esm_parser.choose_blocks(reduced_config)
        esm_parser.recursive_run_function(
            [],
            reduced_config,
            "atomic",
            esm_parser.find_variable,
            reduced_config,
            blacklist,
            True,
        )

        new_config = {}
        for headline in reduced_config:
            if "kind" in reduced_config[headline]:
                if not reduced_config[headline]["kind"] in new_config:
                    new_config[reduced_config[headline]["kind"]] = {
                        headline: reduced_config[headline]
                    }
                else:
                    new_config[reduced_config[headline]["kind"]].update(
                        {headline: reduced_config[headline]}
                    )
            else:
                new_config.update({headline: reduced_config[headline]})

        # esm_parser.pprint_config(new_config)
        # sys.exit(0)
        return new_config

    def replace_last_vars(self, env):

        self.config["computer"] = copy.deepcopy(env.config)
        esm_parser.recursive_run_function(
            [],
            self.config,
            "atomic",
            esm_parser.find_variable,
            self.config,
            [],
            True,
        )

    def update_relevant_entries_with_config(self, config):
        for component in config:
            for entry in self.relevant_entries:
                if (
                    entry in config[component]
                    and component in self.config["components"]
                ):
                    self.config["components"][component][entry] = config[component][
                        entry
                    ]

    def update_packages(self, vcs, general):
        for package in self.all_packages:
            package.fill_in_infos(self, vcs, general)

    def list_all_packages(self, vcs, general):
        packages = []
        config = self.config
        for kind in self.model_kinds:
            for model in config[kind]:
                version = None
                if "available_versions" in config[kind][model]:
                    for version in config[kind][model]["available_versions"]:
                        packages.append(
                            software_package(
                                (kind, model, version),
                                self,
                                vcs,
                                general,
                                no_infos=True,
                            )
                        )
                else:
                    packages.append(
                        software_package(
                            (kind, model, version), self, vcs, general, no_infos=True
                        )
                    )
        return packages

    def has_target(self, package, target, vcs):
        if target in self.meta_todos:
            for subtarget in self.meta_command_order[target]:
                if self.has_target(package, subtarget, vcs):
                    return True
        if target in vcs.known_todos:
            for repo in vcs.known_repos:
                answer = self.get_config_entry(package, repo + "-repository")
                if answer:
                    return True
        else:
            answer = self.get_config_entry(package, target + "_command")
            if answer:
                return True
        return False

    def has_target2(self, package, target):
        for testpackage in self.all_packages:
            if (
                testpackage.raw_name == package.raw_name
                and target in testpackage.targets
            ):
                return True
        return False

    def has_package(self, package):
        if package in self.all_packages:
            return True
        else:
            return False

    def has_model(self, model):
        for kind in self.model_kinds:
            for test_model in self.config[kind]:
                if test_model == model:
                    return True
        return False

    def split_raw_target(self, rawtarget, setup_info):
        todo = kind = only_subtarget = None
        model = version = ""
        if "/" in rawtarget:
            rawtarget, only_subtarget = rawtarget.rsplit("/", 1)

        raw = rawtarget
        for this_todo in setup_info.known_todos:
            if rawtarget == this_todo:
                return this_todo, None, None, None, None, raw
            elif rawtarget.startswith(this_todo + "-"):
                todo = this_todo
                rawtarget = rawtarget.replace(todo + "-", "")
                break

        for package in self.all_packages:
            if package.raw_name == rawtarget:
                return (
                    todo,
                    package.kind,
                    package.model,
                    package.version,
                    only_subtarget,
                    raw,
                )

        # package not found:
        self.output_available_targets(rawtarget)
        sys.exit(0)

    def assemble_raw_name(self, todo, kind, model, version):
        raw = sep = ""
        if todo:
            raw = todo
            sep = "-"
        if model:
            raw = raw + sep + model
            sep = "-"
        if version:
            raw = raw + sep + version
            sep = "-"
        if not raw == "":
            return raw
        return None

    def setup_or_model(self, model):
        kind_of_model = "unknown"
        for kind in self.model_kinds:
            if model in self.config[kind]:
                kind_of_model = kind
        return kind_of_model

    def output_available_targets(self, search_keyword):
        colorama.init(autoreset=True)
        display_info = []
        if search_keyword == "":
            display_info = self.all_packages
        else:
            for package in self.all_packages:
                if package.targets:
                    for target in package.targets:
                        if search_keyword in target + "-" + package.raw_name:
                            if package not in display_info:
                                display_info.append(package)

        if display_info == []:
            print()
            print(
                "No targets found for keyword "
                + search_keyword
                + ". Type 'esm_master' to get a full list"
            )
            print("of available targets.")
            print()
        elif display_info == self.all_packages:
            print()
            print(
                colorama.Fore.YELLOW
                + "Master Tool for ESM applications, including download and compiler wrapper functions"
            )
            print(
                colorama.Fore.YELLOW
                + "		originally written by Dirk Barbi (dirk.barbi@awi.de)"
            )
            print(
                colorama.Fore.YELLOW
                + "       further developed as OpenSource, coordinated and maintained at AWI"
            )
            print()
            print(
                colorama.Fore.YELLOW
                + "Obtain from:         https://github.com/esm-tools/esm_master.git"
            )
            print()
            self.print_nicely(display_info)
            print()
        else:
            print()
            print(
                search_keyword
                + " is not an available target. Did you mean one of those:"
            )
            self.print_nicely(display_info)
            print()

    def print_nicely(self, display_info):
        """Will display all supported models when esm_master is run without
        any arguments or for the selected model

        Parameters
        ----------
        display_info : list of `software_package` objects

        """
        all_available_options = set()  # eg. get, install, clean, ...
        sorted_display = {}
        for kind in self.display_kinds:
            for package in display_info:
                # add the targets of the current package to all targets
                all_available_options.update(package.targets)

                if package.kind == kind:
                    if not kind in sorted_display.keys():
                        sorted_display[kind] = dict()

                    # if model is not encountered yet, then create it first
                    if not package.model in sorted_display[kind]:
                        sorted_display[kind][package.model] = dict()

                    # then add the version : targets node
                    if package.version:
                        sorted_display[kind][package.model][
                            package.version
                        ] = package.targets
                    else:
                        sorted_display[kind][package.model] = package.targets

        # ===
        # print the available models and versions
        # ===
        all_available_options = sorted(all_available_options)
        # sort the dictionary alphabetically for better user experience
        # kind is 'setups', 'components'
        # Each item of the sorted_display[kind].items() is a tuple of "model"
        # and dictionary of version and available options. Sort alphabetically
        # according to the model name, ie. first variable of the tuple
        alphabetical_dict = {}
        for kind in sorted_display:
            # sort the models within each `kind`
            d = dict(
                sorted(sorted_display[kind].items(), key=lambda item: item[0].lower())
            )
            alphabetical_dict[kind] = d

            # sort the versions within each `model`
            for model in alphabetical_dict[kind]:
                # only if it has version : target node, ie omit the models
                # without version
                if isinstance(alphabetical_dict[kind][model], dict):
                    d2 = dict(
                        sorted(
                            alphabetical_dict[kind][model].items(),
                            key=lambda item: item[0].lower(),
                        )
                    )
                    alphabetical_dict[kind][model] = d2

        colorama.init(autoreset=True)
        for kind in alphabetical_dict:
            print(f"{colorama.Fore.GREEN}{kind.upper()}:")
            for model in alphabetical_dict[kind]:
                print(f"    {colorama.Fore.CYAN}{model}:")
                # some models (eg. setups -> oifsamip don't have a version)
                if isinstance(alphabetical_dict[kind][model], list):
                    targets = alphabetical_dict[kind][model]
                    print(f"        ", end="")
                    print(*targets, sep="  ")

                # for the rest of the models [kind][model] will further yield
                # version : targets
                elif isinstance(alphabetical_dict[kind][model], dict):
                    max_version_length = max(
                        [len(ver) for ver in alphabetical_dict[kind][model]]
                    )

                    for version in alphabetical_dict[kind][model]:
                        targets = alphabetical_dict[kind][model][version]
                        print(
                            f"{colorama.Fore.MAGENTA}        "
                            f"{version:<{max_version_length}} :",
                            end=" ",
                        )
                        print(*targets, sep="  ")
            print()

    def get_config_entry(self, package, entry):
        try:
            answer = self.config[package.kind][package.model]["choose_version"][
                package.version
            ][entry]
        except:
            try:
                answer = self.config[package.kind][package.model][entry]
            except:
                answer = None
        return answer

    def output(self):
        print()
        print("Model kinds: " + str(self.model_kinds))
        print("Model todos: " + str(self.model_todos))
        print("All known todos: " + str(self.known_todos))
        for package in self.all_packages:
            package.output()



from . import database
from datetime import datetime


def database_entry(config, action, setup_name, base_dir):
    if config.get("general", {}).get("use_database", True):
        thisrun = (
            database.session.query(database.installation)
            .filter_by(setup_name=setup_name)
            .filter_by(folder=base_dir + "/" + setup_name)
            .all()
        )

        if thisrun == []:
            thisrun = database.installation(
                action=action,
                setup_name=setup_name,
                timestamp=datetime.now(),
                folder=base_dir + "/" + setup_name,
            )
            database.session.add(thisrun)
        else:
            thisrun = thisrun[-1]
            thisrun.action = action
            thisrun.timestamp = datetime.now()
        database.session.commit()
        return database



from . import helpers


def run_job(config):
    config["general"]["relevant_filetypes"] = [
        "log",
        "mon",
        "outdata",
        "restart_out",
        "bin",
        "config",
        "forcing",
        "input",
        "restart_in",
        "ignore",
    ]
    helpers.evaluate(config, "postprocess", "post_recipe")
    return config


def _assemble_postprocess_tasks(config):
    """
    Generates all tasks for post processing which will be written to the run file.

    Parameters
    ----------
    post_file
        File handle to which information should be written.

    Returns
    -------
    post_task_list : list
        The list of post commands which will be executed. These are written
        to the run file.
    """
    postfile = config["general"]["post_file"]

    post_task_list = []
    for component in config["general"]["valid_model_names"]:
        post_file.write(40 * "+ " + "\n")
        post_file.write("Generating post-processing tasks for: %s \n" % component)

        post_task_list.append("\n#Postprocessing %s\n" % component)
        post_task_list.append(
            "cd " + config[component]["experiment_outdata_dir"] + "\n"
        )

        pconfig_tasks = config[component].get("postprocess_tasks", {})
        post_file.write("Configuration for post processing: %s \n" % pconfig_tasks)
        for outfile in pconfig_tasks:
            post_file.write("Generating task to create: %s \n" % outfile)
            ofile_config = pconfig_tasks[outfile]
            # TODO(PG): This can be cleaned up. I probably actually want a
            # ChainMap here for more than just the bottom...
            #
            # Run CDO tasks (default)
            task_definition = (
                config[component]
                .get("postprocess_task_definitions", {})
                .get(ofile_config["post_process"])
            )
            method_definition = (
                config[component]
                .get("postprocess_method_definitions", {})
                .get(task_definition["method"])
            )

            program = method_definition.get("program", task_definition["method"])

            possible_args = method_definition.get("possible_args", [])
            required_args = method_definition.get("required_args", [])

            possible_flags = method_definition.get("possible_flags", [])
            required_flags = method_definition.get("required_flags", [])

            outfile_flags = ofile_config.get("flags")
            outfile_args = ofile_config.get("args")

            task_def_flags = task_definition.get("flags")
            task_def_args = task_definition.get("args")

            args = collections.ChainMap(outfile_args, task_def_args)
            flags = outfile_flags + task_def_flags
            flags = ["-" + flag for flag in flags]

            # See here: https://stackoverflow.com/questions/21773866/how-to-sort-a-dictionary-based-on-a-list-in-python
            all_call_things = {
                "program": program,
                "outfile": outfile,
                **args,
                "flags": flags,
            }
            print(all_call_things)
            index_map = {v: i for i, v in enumerate(method_definition["call_order"])}
            call_list = sorted(
                all_call_things.items(), key=lambda pair: index_map[pair[0]]
            )
            call = []
            for call_id, call_part in call_list:
                if isinstance(call_part, str):
                    call.append(call_part)
                elif isinstance(call_part, list):
                    call.append(" ".join(call_part))
                else:
                    raise TypeError(
                        "Something straaaange happened. Consider starting the debugger."
                    )
            post_file.write(" ".join(call) + "\n")
            post_task_list.append(" ".join(call))
        post_task_list.append("cd -\n")
        config["general"]["post_task_list"] = post_task_list
    return config


# ?????
# def write_simple_postscript(config):
#    batch_system.write_simple_runscript(config)
#    return config



#!/usr/bin/env python
"""
A small wrapper that combines the shell interface and the Python interface
"""

# Import from Python Standard Library
import argparse
import logging
import os
import sys

# Import from 3rd Party packages
import coloredlogs

from .esm_database import *


def parse_shargs():
    """The arg parser for interactive use"""
    parser = argparse.ArgumentParser()
    parser.add_argument("table", nargs="?", default=None)

    parser.add_argument("-f", "--find", help="Find keyword", default=None)

    return parser.parse_args()


def main():
    ARGS = parse_shargs()

    find = None

    parsed_args = vars(ARGS)

    if "find" in parsed_args:
        find = parsed_args["find"]

    db = DisplayDatabase(ARGS.table)



""" ESM Framework for organizing python code in plugins / yaml recipes """

import sys
import esm_parser
from esm_parser import yaml_file_to_dict


def read_recipe(recipe, additional_dict, needs_parse=True):
    if needs_parse:
        recipe = yaml_file_to_dict(recipe)
    recipe.update(additional_dict)
    esm_parser.basic_choose_blocks(recipe, recipe)
    esm_parser.recursive_run_function(
        [], recipe, "atomic", esm_parser.find_variable, recipe, [], True
    )

    return recipe


def read_plugin_information(plugins_bare, recipe, needs_parse=True):
    # pluginfile = esm_plugins.yaml
    if needs_parse:
        plugins_bare = yaml_file_to_dict(plugins_bare)
    extra_info = ["location", "git-url"]
    plugins = {}
    for workitem in recipe["recipe"]:
        found = False
        for module_type in ["core", "plugins"]:
            if module_type in plugins_bare:
                for module in plugins_bare[module_type]:
                    for submodule in plugins_bare[module_type][module]:
                        if submodule in extra_info:
                            continue
                        functionlist = plugins_bare[module_type][module][submodule]
                        if workitem in functionlist:
                            plugins[workitem] = {
                                "module": module,
                                "submodule": submodule,
                                "type": module_type,
                            }
                            for extra in extra_info:
                                if extra in plugins_bare[module_type][module]:
                                    plugins[workitem].update(
                                        {
                                            extra: plugins_bare[module_type][module][
                                                extra
                                            ]
                                        }
                                    )
                            found = True
                            break
                    if found:
                        break
                if found:
                    break
            if found:
                break

    attach_installed_plugins_to_all(plugins)
    return plugins


def find_installed_plugins():
    import pkg_resources

    discovered_plugins = {
        entry_point.name: {"callable": entry_point.load(), "type": "installed"}
        for entry_point in pkg_resources.iter_entry_points("esm_tools.plugins")
    }
    return discovered_plugins


def attach_installed_plugins_to_all(plugins):
    plugins.update(find_installed_plugins())


def check_plugin_availability(plugins):
    something_missing = False
    for workitem in list(plugins.keys()):
        if plugins[workitem]["type"] == "core":
            pass
        elif plugins[workitem]["type"] == "installed":
            pass
        else:
            print(
                "Checking if function "
                + plugins[workitem]["module"]
                + "."
                + plugins[workitem]["submodule"]
                + "."
                + workitem
                + " can be imported..."
            )
            try:
                if sys.version_info >= (3, 5):
                    import importlib.util

                    spec = importlib.util.spec_from_file_location(
                        plugins[workitem]["module"],
                        plugins[workitem]["location"]
                        + "/"
                        + plugins[workitem]["module"]
                        + ".py",
                    )
                    thismodule = importlib.util.module_from_spec(spec)
                    spec.loader.exec_module(thismodule)
            except:
                print(
                    "Couldn't import "
                    + plugins[workitem]["module"]
                    + " from "
                    + plugins[workitem]["location"]
                )
                something_missing = True
    if something_missing:
        sys.exit(-1)


def work_through_recipe(recipe, plugins, config):
    if config.get("general", {}).get("debug_recipe", False):
        import pdb

        pdb.set_trace()
    recipes = recipe["recipe"]
    for index, workitem in enumerate(recipes, start=1):
        if config["general"].get("verbose", False):
            # diagnostic message of which recipe step is being executed
            message = (
                f"::: Executing the step:  {workitem}    "
                f"(step [{index}/{len(recipes)}] of the job:  "
                f'{recipe["job_type"]})'
            )

            print()
            print("=" * len(message))
            print(message)
            print("=" * len(message))
        if plugins[workitem]["type"] == "core":
            thismodule = __import__(plugins[workitem]["module"])
            submodule = getattr(thismodule, plugins[workitem]["submodule"])
            config = getattr(submodule, workitem)(config)
        elif plugins[workitem]["type"] == "installed":
            # print("Installed plugin will be run: ", workitem)
            config = plugins[workitem]["callable"](config)
        else:
            if sys.version_info >= (3, 5):
                import importlib.util

                spec = importlib.util.spec_from_file_location(
                    plugins[workitem]["module"],
                    plugins[workitem]["location"]
                    + "/"
                    + plugins[workitem]["module"]
                    + ".py",
                )
                thismodule = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(thismodule)
                config = getattr(thismodule, workitem)(config)
    return config



"""
Contains functions for dealing with SLURM-based batch systems
"""
import os
import shutil
import subprocess
import sys

import esm_parser

class Slurm:
    """
    Deals with SLURM, allowing you to check if a job is submitted, get the
    current job ID, generate a srun hostfile, get the current job state, and
    check if a job is still running.


    Attributes
    ----------
    filename : str
        The filename for srun commands, defaults to ``hostfile_srun``
    path : str
        Full path to this file, defaults to ``thisrun_scripts_dir / filename``

    Parameters
    ----------
    config : dict
        The run configuration, needed to determine where the script directory
        for this particular run is.
    """

    def __init__(self, config):
        folder = config["general"]["thisrun_scripts_dir"]
        self.filename = "hostfile_srun"
        self.path = folder + "/" + self.filename

    @staticmethod
    def check_if_submitted():
        """
        Determines if a job is submitted in the currently running shell by
        checking for ``SLURM_JOB_ID`` in the environment

        Returns
        -------
        bool
        """
        return "SLURM_JOB_ID" in os.environ

    @staticmethod
    def get_jobid():
        """
        Gets the current SLURM JOB ID

        Returns
        -------
        str or None
        """
        return os.environ.get("SLURM_JOB_ID")

    def prepare_launcher(self, config, cluster):
        # MA: not sure how this will play with heterogeneous parallelization
        if "multi_srun" in config["general"]:
            for run_type in list(config["general"]["multi_srun"]):
                current_hostfile = self.path + "_" + run_type
                write_one_hostfile(current_hostfile, config)

        if (
            config["computer"].get("heterogeneous_parallelization", False)
            and not config["computer"].get("taskset", False)
        ):
            # Prepare heterogeneous parallelization call
            config["general"]["batch"].het_par_launcher_lines(config, cluster)
        else:
            # Standard/old way of running jobs with slurm
            self.write_one_hostfile(self.path, config)

            hostfile_in_work = (
                config["general"]["work_dir"] + "/" + os.path.basename(self.path)
            )
            shutil.copyfile(self.path, hostfile_in_work)

        return config

    def write_one_hostfile(self, hostfile, config):
        """
        Gathers previously prepared requirements
        (batch_system.calculate_requirements) and writes them to ``self.path``.
        """

        with open(hostfile, "w") as hostfile:
            for model in config["general"]["valid_model_names"]:
                end_proc = config[model].get("end_proc", None)
                start_proc = config[model].get("start_proc", None)

                if start_proc == None or end_proc == None:
                    continue

                if config["computer"].get("heterogeneous_parallelization", False):
                    command = "./" + config[model].get(
                        "execution_command_het_par", None
                    )
                elif "execution_command" in config[model]:
                    command = "./" + config[model]["execution_command"]
                elif "executable" in config[model]:
                    command = "./" + config[model]["executable"]
                else:
                    continue
                hostfile.write(
                    str(start_proc) + "-" + str(end_proc) + "  " + command + "\n"
                )

    @staticmethod
    def get_job_state(jobid):
        """
        Returns the jobstate full name. See ``man squeue``, section ``JOB STATE CODES`` for more details.

        Parameters
        ----------
        jobid :
            ``str`` or ``int``. The SLURM job id as displayed in, e.g. ``squeue``

        Returns
        -------
        str :
            The short job state.
        """
        state_command = ["squeue -j" + str(jobid) + ' -o "%T"']

        squeue_output = subprocess.Popen(
            state_command, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        ).communicate()[0]
        if len(squeue_output) == 2:
            return squeue_output[0]

    @staticmethod
    def job_is_still_running(jobid):
        """Returns a boolean if the job is still running"""
        return bool(Slurm.get_job_state(jobid))

    ############# HETEROGENOUS PARALLELIZATION STUFF (MPI + OMP) #################

    def add_pre_launcher_lines(self, config, cluster, runfile):
        """
        Adds pre-launcher lines to the ``runfile``.

        Parameters
        ----------
        config : dict
            Configuration dictionary containing information about the experiment and
            experiment directory.
        runfile : io.TextIOWrapper
            File wrapper object for writing of the lines
            (``runfile.write("<your_line_here>")``).
        """

        # TODO: remove it once it's not needed anymore (substituted by packjob)
        if config["computer"].get("heterogeneous_parallelization", False):
            if config["computer"].get("taskset", False):
                self.add_hostlist_file_gen_lines(config, runfile)

    @staticmethod
    def het_par_headers(config, cluster, headers):
        """
        Modifies the list of ``headers`` to include the ``packjob``/``hetjob`` logic
        for heterogeneous parallelization in SLURM.

        Parameters
        ----------
        config : dict
            Configuration dictionary containing information about the experiment and
            experiment directory.
        cluster : str
            Type of job cluster.
        headers : list
            List of headers for the ``.run`` file.

        Returns
        -------
        headers : list
            List of headers for the ``.run`` file, with the necessary modifications
            for heterogeneous parallelization in SLURM.
        """
        # Only modify the headers if ``heterogeneous_parallelization`` is ``True``
        if (
            config["computer"].get("heterogeneous_parallelization", False)
            and not config["computer"].get("taskset", False)
        ):
            this_batch_system = config["computer"]
            # Get the variables to be modified for the headers
            nodes_flag = this_batch_system["nodes_flag"].split("=")[0]
            partition_flag = this_batch_system["partition_flag"].split("=")[0]
            # Delete ``nodes_flag`` and ``partition_flag`` from the ``headers``. Those
            # will be added at the end of the headers for each component
            headers_new = []
            for val in headers:
                 if nodes_flag not in val and partition_flag not in val:
                    headers_new.append(val)
            headers = headers_new
            # Loop through the models to add the respective component headers (nodes
            # and partitions)
            for model in config["general"]["valid_model_names"]:
                nodes = config[model].get("nodes")

# kh 24.06.22 observed behavior was:
#model:  echam, nodes:  4
#model:  fesom, nodes:  32
#model:  oasis3mct, nodes:  None
#model:  recom, nodes:  None
#model:  jsbach, nodes:  None
#model:  hdmodel, nodes:  1

# kh 24.06.22 workaround: filter hdmodel
                if nodes:
                    headers.append(f"{nodes_flag}={nodes}")
                    headers.append(this_batch_system["partition_flag"])
                    headers.append(this_batch_system["hetjob_flag"])
            # Remove the last hetjob_flag
            if headers[-1] == this_batch_system["hetjob_flag"]:
                headers = headers[:-1]

        return headers

    # TODO: remove it once it's not needed anymore (substituted by packjob)
    @staticmethod
    def write_het_par_wrappers(config):
        cores_per_node = config["computer"]["partitions"]["compute"]["cores_per_node"]

        scriptfolder = config["general"]["thisrun_scripts_dir"] + "../work/"
        if config["computer"].get("heterogeneous_parallelization", False):
            for model in config["general"]["valid_model_names"]:
                if "oasis3mct" == model:
                    continue
                elif (
                        not config[model].get("execution_command")
                        and not config[model].get("executable")
                ):
                    esm_parser.user_note(
                        "Execution command",
                        f"Execution command for ``{model}`` not found. This is okay " \
                        "if this component has no binary to be called, but if it does" \
                        " please make sure you specify either an ``executable`` " \
                        f"or an ``execution_command`` variable in the ``{model}`` " \
                        "of your configuration file.",
                    )
                    continue

                command = "./" + config[model].get(
                    "execution_command", config[model]["executable"]
                )
                scriptname = "script_" + model + ".sh"
                with open(scriptfolder + scriptname, "w") as f:
                    f.write("#!/bin/sh" + "\n")
                    f.write(
                        "export OMP_NUM_THREADS="
                        + str(config[model].get("omp_num_threads", 1))
                        + "\n"
                    )
                    f.write(command + "\n")
                os.chmod(scriptfolder + scriptname, 0o755)

                progname = "prog_" + model + ".sh"

                start_core = config[model]["start_core"]
                end_core = config[model]["end_core"]

                with open(scriptfolder + progname, "w") as f:
                    f.write("#!/bin/sh" + "\n")
                    f.write("if [ -z ${PMI_RANK+x} ]; then PMI_RANK=$PMIX_RANK; fi" + "\n")
                    f.write("(( init = $PMI_RANK ))" + "\n")
                    f.write(
                        "(( index = init * "
                        + str(config[model].get("omp_num_threads", 1))
                        + " ))"
                        + "\n"
                    )
                    f.write("(( slot = index % " + str(cores_per_node) + " ))" + "\n")
                    f.write(
                        "echo "
                        + model
                        + " taskset -c $slot-$((slot + "
                        + str(config[model].get("omp_num_threads", 1))
                        + " - 1"
                        + "))"
                        + "\n"
                    )
                    f.write(
                        "taskset -c $slot-$((slot + "
                        + str(config[model].get("omp_num_threads", 1))
                        + " - 1)) ./script_"
                        + model
                        + ".sh"
                        + "\n"
                    )
                os.chmod(scriptfolder + progname, 0o755)
                execution_command_het_par = f"prog_{model}.sh"
                config[model]["execution_command_het_par"] = execution_command_het_par
        return config

    # TODO: remove it once it's not needed anymore (substituted by packjob)
    @staticmethod
    def add_hostlist_file_gen_lines(config, runfile):
        cores_per_node = config["computer"]["partitions"]["compute"]["cores_per_node"]
        runfile.write(
            "\n"
            + "#Creating hostlist for MPI + MPI&OMP heterogeneous parallel job"
            + "\n"
        )
        runfile.write("rm -f ./hostlist" + "\n")
        runfile.write(
            f"export SLURM_HOSTFILE={config['general']['thisrun_work_dir']}/hostlist\n"
        )
        runfile.write("IFS=$'\\n'; set -f" + "\n")
        runfile.write(
            "listnodes=($(< <( scontrol show hostnames $SLURM_JOB_NODELIST )))" + "\n"
        )
        runfile.write("unset IFS; set +f" + "\n")
        runfile.write("rank=0" + "\n")
        runfile.write("current_core=0" + "\n")
        runfile.write("current_core_mpi=0" + "\n")
        for model in config["general"]["valid_model_names"]:
            if (
                "oasis3mct" != model
                and (
                    config[model].get("execution_command")
                    or config[model].get("executable")
                )
            ):
                if "nproca" in config[model]:
                    mpi_tasks = config[model]["nproca"] * config[model]["nprocb"]
                else:
                    mpi_tasks = config[model]["nproc"]
                runfile.write(
                    "mpi_tasks_" + model + "=" + str(mpi_tasks) + "\n"
                )
                runfile.write(
                    "omp_threads_"
                    + model
                    + "="
                    + str(config[model].get("omp_num_threads", 1))
                    + "\n"
                )
        import pdb

        # pdb.set_trace()
        runfile.write(
            "for model in "
            + str(config["general"]["valid_model_names"])[1:-1]
            .replace(",", "")
            .replace("'", "")
            + " ;do"
            + "\n"
        )
        runfile.write("    eval nb_of_cores=\${mpi_tasks_${model}}" + "\n")
        runfile.write("    eval nb_of_cores=$((${nb_of_cores}-1))" + "\n")
        runfile.write("    for nb_proc_mpi in `seq 0 ${nb_of_cores}`; do" + "\n")
        runfile.write(
            "        (( index_host = current_core / "
            + str(cores_per_node)
            + " ))"
            + "\n"
        )
        runfile.write("        host_value=${listnodes[${index_host}]}" + "\n")
        runfile.write(
            "        (( slot =  current_core % " + str(cores_per_node) + " ))" + "\n"
        )
        runfile.write("        echo $host_value >> hostlist" + "\n")
        runfile.write(
            "        (( current_core = current_core + omp_threads_${model} ))" + "\n"
        )
        runfile.write("    done" + "\n")
        runfile.write("done" + "\n\n")

    ############# MULTI SRUN STUFF ##############

    @staticmethod
    def determine_nodelist(config):
        setup_name = config["general"]["setup_name"]
        if config["general"].get("multi_srun"):
            for run_type in config["general"]["multi_srun"]:
                print(run_type)
                total_tasks = 0
                for model in config["general"]["multi_srun"][run_type]["models"]:
                    print(total_tasks)
                    # determine how many nodes that component needs
                    if "nproc" in config[model]:
                        print("Adding to total_tasks")
                        total_tasks += int(config[model]["nproc"])
                        print(total_tasks)
                    elif "nproca" in config[model] and "nprocb" in config[model]:
                        print("Adding to total_tasks")
                        total_tasks += int(config[model]["nproca"]) * int(
                            config[model]["nprocb"]
                        )
                        print(total_tasks)

                        # KH 30.04.20: nprocrad is replaced by more flexible
                        # partitioning using nprocar and nprocbr
                        if "nprocar" in config[model] and "nprocbr" in config[model]:
                            if (
                                config[model]["nprocar"] != "remove_from_namelist"
                                and config[model]["nprocbr"] != "remove_from_namelist"
                            ):
                                print("Adding to total_tasks")
                                total_tasks += (
                                    config[model]["nprocar"] * config[model]["nprocbr"]
                                )
                                print(total_tasks)

                    else:
                        continue
                config["general"]["multi_srun"][run_type]["total_tasks"] = total_tasks
            print(config["general"]["multi_srun"])
        return config


def get_run_commands_multisrun(config, commands):
    default_exec_command = config["computer"]["execution_command"]
    print("---> This is a multi-srun job.")
    print("The default command:")
    print(default_exec_command)
    print("Will be replaced")
    # Since I am already confused, I need to write comments.
    #
    # The next part is actually a shell script fragment, which will be injected
    # into the "run" file, the shell script to be run by the job scheduler.
    #
    # In this part, we figure out what compute nodes we are using so we can
    # specify nodes for each srun command. That means, ECHAM+FESOM will use one
    # pre-defined set of nodes, PISM another, and so on. That should be general
    # enough to also work for other model combos...
    #
    # Not sure if this is specific to Mistral as a HPC, Slurm as a batch
    # system, or whatever else might pop up...
    # @Dirk, please move this where you see it best (I guess slurm.py)
    job_node_extraction = r"""
    # Job nodes extraction
    nodeslurm=$SLURM_JOB_NODELIST
    echo "nodeslurm = ${nodeslurm}"
    # Get rid of the hostname and surrounding brackets:
    tmp=${nodeslurm#"*["}
    nodes=${tmp%]*}
    # Turn it into an array seperated by newlines:
    myarray=(`echo ${nodes} | sed 's/,/\n/g'`)
    #
    idx=0
    for element in "${myarray[@]}"; do
        if [[ "$element" == *"-"* ]]; then
            array=(`echo $element | sed 's/-/\n/g'`)
            for node in $(seq ${array[0]} ${array[1]}); do
               nodelist[$idx]=${node}
               idx=${idx}+1
            done
        else
            nodelist[$idx]=${element}
            idx=${idx}+1
        fi
    done

    for element in "${nodelist[@]}"; do
        echo "${element}"
    done
    """

    def assign_nodes(
        run_type, need_length=False, start_node=0, num_nodes_first_model=0
    ):
        template = f"""
        # Assign nodes for {run_type}
        {run_type}=""
        %%NEED_LENGTH%%
        for idx in $srbseq {start_node} $srbsrb???-1erberberb; do
            if ssbssb $idx == $srbsrb???-1erberb esbesb; then
                {run_type}="$scb{run_type}ecb$scbnodelist[$idx]ecb"
            else
                {run_type}="$scb{run_type}ecb$scbnodelistssb$idxesbecb,"
            fi
        done
        echo "{run_type} nodes: $scb{run_type}ecb"
        """
        # Since Python f-strings and other braces don't play nicely together,
        # we replace some stuff:
        #
        # For the confused:
        # scb = start curly brace {
        # ecb = end curly brace }
        # ssb = start square brace [
        # esb = end square brace ]
        # srb = start round brace (
        # erb = end round brace )
        template = template.replace("scb", "{")
        template = template.replace("ecb", "}")
        template = template.replace("ssb", "[")
        template = template.replace("esb", "]")
        template = template.replace("srb", "(")
        template = template.replace("erb", ")")
        # Get rid of the starting spaces (they come from Python as the string
        # is defined inside of this function which is indented (facepalm))
        template = textwrap.dedent(template)
        # TODO: Some replacements
        if need_length:
            length_stuff = r"length=${#nodelist[@]}"
            template = template.replace("%%NEED_LENGTH%%", length_stuff)
            template = template.replace("???", "length")
        else:
            template = template.replace("%%NEED_LENGTH%%", "")
            template = template.replace("???", str(num_nodes_first_model))
        return template

    commands.append(textwrap.dedent(job_node_extraction))
    cores_per_node = config["computer"]["partitions"]["compute"]["cores_per_node"]
    for idx, run_type in enumerate(config["general"]["multi_srun"]):
        if idx == 0:
            start_node = run_type
            num_nodes_first_model = (
                config["general"]["multi_srun"][run_type]["total_tasks"]
                / cores_per_node
            )
            num_nodes_first_model = int(num_nodes_first_model)
            nodes = assign_nodes(
                run_type, need_length=False, num_nodes_first_model=num_nodes_first_model
            )
        else:
            nodes = assign_nodes(run_type, need_length=True, start_node=start_node)
        commands.append(nodes)
    for run_type in config["general"]["multi_srun"]:
        new_exec_command = default_exec_command.replace(
            "hostfile_srun", config["general"]["multi_srun"][run_type]["hostfile"]
        )
        new_exec_command += f" --nodelist ${run_type}"
        commands.append("time " + new_exec_command + " &")
    return commands



# -*- coding: utf-8 -*-

"""Top-level package for ESM Archiving."""

__author__ = """Paul Gierz"""
__email__ = "pgierz@awi.de"
__version__ = "6.17.5"

from .esm_archiving import (archive_mistral, check_tar_lists,
                            delete_original_data, determine_datestamp_location,
                            determine_potential_datestamp_locations,
                            find_indices_of, get_files_for_date_range,
                            get_list_from_filepattern, group_files,
                            group_indexes, log_tarfile_contents, pack_tarfile,
                            purify_expid_in, sort_files_to_tarlists,
                            split_list_due_to_size_limit, stamp_filepattern,
                            stamp_files, sum_tar_lists,
                            sum_tar_lists_human_readable)

__all__ = (
    "archive_mistral",
    "check_tar_lists",
    "delete_original_data",
    "determine_datestamp_location",
    "determine_potential_datestamp_locations",
    "find_indices_of",
    "get_files_for_date_range",
    "get_list_from_filepattern",
    "group_files",
    "group_indexes",
    "log_tarfile_contents",
    "pack_tarfile",
    "purify_expid_in",
    "sort_files_to_tarlists",
    "split_list_due_to_size_limit",
    "stamp_filepattern",
    "stamp_files",
    "sum_tar_lists",
    "sum_tar_lists_human_readable",
)



import os
import sys

import esm_parser
import esm_tools

######################################################################################
##################################### globals ########################################
######################################################################################

ESM_MASTER_DIR = os.getenv("PWD")

COMPONENTS_DIR = esm_tools.get_config_filepath("/components/")
SETUPS_DIR = esm_tools.get_config_filepath("/setups/")
COUPLINGS_DIR = esm_tools.get_config_filepath("/couplings/")
DEFAULTS_DIR = esm_tools.get_config_filepath("/defaults/")
ESM_SOFTWARE_DIR = esm_tools.get_config_filepath("/esm_software/")
CONFIG_YAML = esm_tools.get_config_filepath("/esm_software/esm_master/esm_master.yaml")
VCS_FOLDER = esm_tools.get_config_filepath("/other_software/vcs/")

ESM_MASTER_PICKLE = ESM_SOFTWARE_DIR + "/esm_master/esm_master.pkl"

######################################################################################
##################################### tab completion #################################
######################################################################################


def tab_completion(parsed_args, setups2models):
    if parsed_args["list_all_targets"]:
        all_commands = []
        for package in setups2models.all_packages:
            for command in package.command_list:
                all_commands.append(command + "-" + package.raw_name)
        print("\n".join(all_commands))
        sys.exit()

    if parsed_args["generate_tab_complete"]:
        with open("esm_master_tabcomplete.bash", "w") as tab_comp:
            tab_comp.write("#/usr/bin/env bash\n")
            tab_comp.write("_esm_master_completions() {\n")
            tab_comp.write(
                '\tCOMPREPLY=($(compgen -W "$(esm_master --list_all_targets)" "${COMP_WORDS[1]}"))'
            )
            tab_comp.write("\n}\n\ncomplete -F _esm_master_completions esm_master\n")
        print("Wrote file: esm_master_tabcomplete.bash")
        print(
            "Have your shell source this file to allow tab completion of available targets"
        )
        print("This works for both bash and zsh")
        return 0


######################################################################################
############################## Write a tiny user script ##############################
######################################################################################


def write_minimal_user_config(config):
    """

    In order to generate a SimulationSetup using esm_parser, we need a small and
    simple user_config. It doesn't really matter for esm_master if all the entries
    are correct, as we don't need most of them, but the esm_parser fails if they
    don't exist as all...
    Needs the name of the model / setup, and version (if exists) as input.

    """
    user_config = {}

    for model in config["components"]:
        version = config["components"][model]["version"]
        model_dir = config["components"][model]["model_dir"]
        user_config.update(
            {model: {"model": model, "version": version, "model_dir": model_dir}}
        )

    if "setups" in config:
        coupled = "true"
        setup = list(config["setups"])[0]
        version = config["setups"][setup]["version"]
        model_dir = config["setups"][setup]["model_dir"]

    else:
        coupled = "false"
        setup = list(config["components"])[0]
        version = config["components"][setup]["version"]

    user_config.update(
        {
            "general": {
                "jobtype": "compute",
                "account": "none",
                "setup_name": setup,
                "version": version,
                "coupled": "false",
                "initial_date": "2000-01-01",
                "final_date": "2000-01-02",
                "compute_time": "00:00:01",
                "model_dir": model_dir,
                "base_dir": "/on/a/road/to/nowwhere",
            }
        }
    )

    if "general.yaml" in os.listdir(DEFAULTS_DIR):
        general_config = esm_parser.yaml_file_to_dict(f"{DEFAULTS_DIR}/general.yaml")
        user_config["general"].update(general_config)

    return user_config


######################################################################################
############################## class "GeneralInfos" ##################################
######################################################################################


class GeneralInfos:
    def __init__(self, parsed_args):

        # Parses the ``esm_master.yaml`` configuration file
        self.config = esm_parser.yaml_file_to_dict(CONFIG_YAML)
        self.meta_todos, self.meta_command_order = self.get_meta_command()
        self.display_kinds = self.get_display_kinds()

        if parsed_args.get("verbose", False):
            self.output()

    def get_meta_command(self):
        """
        Gets ``meta_todos`` and ``meta_command_order``, which are a combination
        of other tasks. e.g. "install" does get, conf, and comp.

        Returns
        -------
        Tuple of List, Dict
        The list contains all meta_todos, the dict contains the todo as the
        key, and the steps as the value.
        """
        meta_todos = []
        meta_command_order = {}
        for entry in self.config:
            if entry.endswith("_meta_command"):
                todo = entry.replace("_meta_command", "")
                meta_todos.append(todo)
                meta_command_order.update({todo: self.config[entry]})
        return meta_todos, meta_command_order

    def get_display_kinds(self):
        if "display_kinds" in self.config.keys():
            return self.config["display_kinds"]

    def output(self):
        print()
        print("Meta commands: ")
        for key in self.meta_command_order:
            print("    " + key + ": " + str(self.meta_command_order[key]))


######################################################################################
########################### class "version_control_infos" ############################
######################################################################################


class version_control_infos:
    def __init__(self, parsed_args):
        self.config = {}
        vcs_files = [f for f in os.listdir(VCS_FOLDER)]
        self.known_repos = []
        for vcs_file in vcs_files:
            if os.path.isfile(VCS_FOLDER + "/" + vcs_file):
                repo_type = vcs_file.replace(".yaml", "")
                self.config.update(
                    {
                        repo_type: esm_parser.yaml_file_to_dict(
                            VCS_FOLDER + "/" + vcs_file
                        )
                    }
                )
                self.known_repos.append(repo_type)
        self.known_todos = []
        for repo in self.known_repos:
            for entry in self.config[repo].keys():
                if entry.endswith("_command"):
                    todo = entry.replace("_command", "")
                    if todo not in self.known_todos:
                        self.known_todos.append(todo)
        if parsed_args.get("verbose", False):
            self.output()

    def assemble_command(self, package, todo, setup_info, general):
        if package.repo_type and package.repo:
            if todo in setup_info.meta_todos:
                return None
            try:
                raw_command = self.config[package.repo_type][todo + "_command"]

                # kh 11.09.20 support git options like --recursive
                # repo_options in the model.yaml is assigned to define_options
                if package.repo_options:
                    define_options = self.config[package.repo_type]["define_options"]
                    raw_command = raw_command.replace(
                        "${define_options}", define_options
                    )
                    raw_command = raw_command.replace(
                        "${repo_options}", package.repo_options
                    )
                else:
                    raw_command = raw_command.replace("${define_options} ", "")
                    raw_command = raw_command.replace(
                        "${repo_options}", ""
                    )  # kh 11.09.20 should not really be necessary
                if package.branch:
                    define_branch = self.config[package.repo_type]["define_branch"]
                    raw_command = raw_command.replace("${define_branch}", define_branch)
                    raw_command = raw_command.replace("${branch}", package.branch)
                else:
                    raw_command = raw_command.replace("${define_branch} ", "")
                    raw_command = raw_command.replace("${branch}", "")

                if package.tag:
                    define_tag = self.config[package.repo_type]["define_tag"]
                    raw_command.replace("${define_tag}", define_tag)
                    raw_command.replace("${tag}", package.tag)
                else:
                    raw_command = raw_command.replace("${define_tag} ", "")
                    raw_command = raw_command.replace("${tag}", "")

                # deniz: pipe support. Eg. curl foo.tar.gz | tar xz
                # pipe_options is given in model yaml file
                if package.pipe_options:
                    raw_command = raw_command.replace(
                        "${pipe_options}", package.pipe_options
                    )
            except:
                print("Sorry, no " + todo + "_command defined for " + package.repo_type)
                sys.exit(42)
            if type(package.repo) == list:
                repo = package.repo[0]
            else:
                repo = package.repo
            if os.environ.get("CI"):
                if "gitlab.awi.de" in repo:
                    awi_user = os.environ.get("GITLAB_AWI_USER_NAME", "")
                    if awi_user:
                        awi_user += "@"
                    repo = repo.replace("gitlab.awi.de", f"{awi_user}gitlab.dkrz.de")
                elif "swrepo1.awi.de" in repo:
                    print(
                        f"swrepo1.awi.de is decommisioned, please consider correcting {repo}"
                    )
                    if os.environ.get("esm_tools_pedantic") or os.environ.get(
                        "ESM_TOOLS_PEDANTIC"
                    ):
                        print("Pedantic mode is on, exiting instead of warning only!")
                        sys.exit(1)
                elif "gitlab.dkrz.de" in repo:
                    # NOTE(PG): Empty string as return value to ensure that the f-string works nicely
                    dkrz_user = os.environ.get("GITLAB_DKRZ_USER_NAME", "")
                    if dkrz_user:
                        dkrz_user += "@"
                    repo = repo.replace("gitlab.dkrz.de", f"{dkrz_user}gitlab.dkrz.de")
                elif "github.com" in repo:
                    print("No token needed for github.com repositories!")
                else:
                    print(f"Sorry, no CI token defined for {repo}")
                    # FIXME(PG): We should come up with a sensible convention for this:
                    if os.environ.get("esm_tools_pedantic") or os.environ.get(
                        "ESM_TOOLS_PEDANTIC"
                    ):
                        print("Pedantic mode is on, exiting instead of warning only!")
                        sys.exit(1)

            raw_command = raw_command.replace("${repository}", repo)
            if todo == "get":
                if package.repo_type == "curl":
                    raw_command = raw_command.replace("${curl-repository}", repo)
                    # return so that it is not overwritten
                    return raw_command
                if package.clone_destination:
                    raw_command = raw_command + " " + package.clone_destination
                elif package.destination:
                    raw_command = raw_command + " " + package.destination
                else:
                    raw_command = raw_command + " " + package.raw_name
        else:
            raw_command = None
        return raw_command

    def output(self):
        print()
        esm_parser.pprint_config(self.config)
        print("Known repos: " + str(self.known_repos))
        print("Known vcs-commands: " + str(self.known_todos))



"""
Functionality for listing plugins via the command line.

Example::

    $ esm_plugins

    Plugin 1
    --------
    Docstring of plugin 1

    Plugin 2
    --------
    Docstring of plugin 2

"""
from esm_plugin_manager import find_installed_plugins


def main():
    """The main plugin CLI"""
    discovered_plugins = find_installed_plugins()
    print("The following plugins are installed and available:")
    for plugin_name in discovered_plugins:
        plugin_code = discovered_plugins[plugin_name]["callable"]
        print(plugin_name)
        doc = plugin_code.__doc__
        if doc:
            print("-" * len(plugin_name))
            print(doc)



import copy
import datetime
import filecmp
import glob
import os
import pathlib
import re
import shutil
import sys
import time

import f90nml
import yaml

import esm_parser
import esm_tools

from . import helpers


def rename_sources_to_targets(config):
    # Purpose of this routine is to make sure that filetype_sources and
    # filetype_targets are set correctly, and _in_work is unset
    for filetype in config["general"]["all_model_filetypes"]:
        for model in config["general"]["valid_model_names"] + ["general"]:

            sources = filetype + "_sources" in config[model]
            targets = filetype + "_targets" in config[model]
            in_work = filetype + "_in_work" in config[model]

            if (
                filetype in config["general"]["out_filetypes"]
            ):  # stuff to be copied out of work

                if sources and targets and in_work:
                    if (
                        not config[model][filetype + "_sources"]
                        == config[model][filetype + "_in_work"]
                    ):
                        print(
                            "Mismatch between "
                            + filetype
                            + "_sources and "
                            + filetype
                            + "_in_work in model "
                            + model,
                            flush=True,
                        )
                        helpers.print_datetime(config)
                        sys.exit(-1)

                elif sources and targets and not in_work:
                    # all fine
                    pass

                elif sources and not targets:
                    if config["general"]["verbose"]:
                        print(
                            "Renaming sources to targets for filetype "
                            + filetype
                            + " in model "
                            + model,
                            flush=True,
                        )
                        helpers.print_datetime(config)
                    config[model][filetype + "_targets"] = copy.deepcopy(
                        config[model][filetype + "_sources"]
                    )
                    if in_work:
                        config[model][filetype + "_sources"] = copy.deepcopy(
                            config[model][filetype + "_in_work"]
                        )

                elif targets and not sources:
                    if in_work:
                        config[model][filetype + "_sources"] = copy.deepcopy(
                            config[model][filetype + "_in_work"]
                        )
                    else:
                        config[model][filetype + "sources"] = copy.deepcopy(
                            config[model][filetype + "_targets"]
                        )

            else:  # stuff to be copied into work

                if sources and targets and in_work:
                    if (
                        not config[model][filetype + "_targets"]
                        == config[model][filetype + "_in_work"]
                    ):
                        print(
                            "Mismatch between "
                            + filetype
                            + "_targets and "
                            + filetype
                            + "_in_work in model "
                            + model,
                            flush=True,
                        )
                        helpers.print_datetime(config)
                        sys.exit(-1)

                elif sources and targets and not in_work:
                    # all fine
                    pass

                elif (not sources and in_work) or (not sources and targets):
                    print(filetype + "_sources missing in model " + model, flush=True)
                    helpers.print_datetime(config)
                    sys.exit(-1)

                elif sources and not targets:
                    if in_work:
                        config[model][filetype + "_targets"] = copy.deepcopy(
                            config[model][filetype + "_in_work"]
                        )
                    else:
                        config[model][filetype + "_targets"] = {}
                        for descrip, name in config[model][filetype + "_sources"].items():
                            config[model][filetype + "_targets"].update(
                                {descrip: os.path.basename(name)}
                            )

            if in_work:
                del config[model][filetype + "_in_work"]

    return config


def complete_targets(config):
    for filetype in config["general"]["all_model_filetypes"]:
        for model in config["general"]["valid_model_names"] + ["general"]:
            if filetype + "_sources" in config[model]:
                for category in config[model][filetype + "_sources"]:
                    if not category in config[model][filetype + "_targets"]:
                        file_source = config[model][filetype + "_sources"][category]

                        # check if the file_source has the correct type. For
                        # unresolved variables they may still be a 'dict'
                        if not isinstance(file_source, (str, os.PathLike)):
                            # model, scenario and version are omitted to make
                            # error message less verbose
                            scenario = config[model].get("scenario", "UNDEFINED")
                            version = config[model].get("version", "UNDEFINED")

                            error_type = "Missing Configuration"
                            error_text = (
                                # comment-out the line below to provide more information
                                # f"Scenario {scenario} for the model {model} (version: {version}) has not been implemented yet. \n" +
                                f"The input file variable {category} of {filetype}_sources can not be fully resolved:\n\n"
                                + yaml.dump(file_source, indent=4)
                            )
                            esm_parser.user_error(error_type, error_text)
                        else:
                            config[model][filetype + "_targets"][
                                category
                            ] = os.path.basename(file_source)

    return config


def complete_sources(config):
    if config["general"]["verbose"]:
        print("::: Complete sources", flush=True)
        helpers.print_datetime(config)
    for filetype in config["general"]["out_filetypes"]:
        for model in config["general"]["valid_model_names"] + ["general"]:
            if filetype + "_sources" in config[model]:
                for category in config[model][filetype + "_sources"]:
                    if not config[model][filetype + "_sources"][category].startswith("/"):
                        config[model][filetype + "_sources"][category] = (
                            config["general"]["thisrun_work_dir"]
                            + "/"
                            + config[model][filetype + "_sources"][category]
                        )
    return config


def reuse_sources(config):
    if config["general"]["run_number"] == 1:
        return config

    # MA: the changes below are to be able to specify model specific reusable_filetypes
    # without changing the looping order (a model loop nested inside a file-type loop)

    # Put together all the possible reusable file types
    all_reusable_filetypes = []
    for model in config["general"]["valid_model_names"] + ["general"]:
        all_reusable_filetypes = list(
            set(all_reusable_filetypes + config[model].get("reusable_filetypes", []))
        )
    # Loop through all the reusable file types
    for filetype in all_reusable_filetypes:
        for model in config["general"]["valid_model_names"] + ["general"]:
            # Get the model-specific reusable_filetypes, if not existing, get the
            # general ones
            model_reusable_filetypes = config[model].get(
                "reusable_filetypes", config["general"]["reusable_filetypes"]
            )
            # Apply changes from ``--update-files`` flag
            model_reusable_filetypes = helpers.update_reusable_filetypes(
                config, reusable_filetypes=model_reusable_filetypes
            )
            # If <filetype>_sources dictionary exists and filetype is in the
            # model-specific filetype list then add the sources
            if (
                filetype + "_sources" in config[model]
                and filetype in model_reusable_filetypes
            ):
                for category in config[model][filetype + "_sources"]:
                    config[model][filetype + "_sources"][category] = (
                        config[model]["experiment_" + filetype + "_dir"]
                        + "/"
                        + config[model][filetype + "_targets"][category]
                    )
    return config


def choose_needed_files(config):
    # aim of this function is to only take those files specified in fileytype_files
    # (if exists), and then remove filetype_files

    for filetype in config["general"]["all_model_filetypes"]:
        for model in config["general"]["valid_model_names"] + ["general"]:

            if not filetype + "_files" in config[model]:
                continue

            new_sources = new_targets = {}
            for category, name in config[model][filetype + "_files"].items():
                # TODO: change with user_error()
                if not name in config[model][filetype + "_sources"]:
                    print(
                        "Implementation "
                        + name
                        + " not found for filetype "
                        + filetype
                        + " of model "
                        + model,
                        flush=True,
                    )
                    print(config[model][filetype + "_files"], flush=True)
                    print(config[model][filetype + "_sources"], flush=True)
                    helpers.print_datetime(config)
                    sys.exit(-1)
                new_sources.update({category: config[model][filetype + "_sources"][name]})

            config[model][filetype + "_sources"] = new_sources

            all_categs = list(config[model][filetype + "_targets"].keys())
            for category in all_categs:
                if not category in config[model][filetype + "_sources"]:
                    del config[model][filetype + "_targets"][category]

            del config[model][filetype + "_files"]

    return config


def globbing(config):
    for filetype in config["general"]["all_model_filetypes"]:
        for model in config["general"]["valid_model_names"] + ["general"]:
            if filetype + "_sources" in config[model]:
                # oldconf = copy.deepcopy(config[model])
                for descr, filename in copy.deepcopy(config[model][filetype + "_sources"]).items():
                    if "*" in filename:
                        del config[model][filetype + "_sources"][descr]
                        # Save the wildcard string for later use when copying to target
                        wild_card = config[model].get(
                            f"{filetype}_sources_wild_card", {}
                        )
                        wild_card[descr] = filename.split("/")[-1]
                        config[model][filetype + "_sources_wild_card"] = wild_card
                        # skip subdirectories in file list, otherwise they
                        # will be listed as missing files later on
                        all_filenames = [
                            f for f in glob.glob(filename) if not os.path.isdir(f)
                        ]
                        running_index = 0

                        for new_filename in all_filenames:
                            newdescr = descr + "_glob_" + str(running_index)
                            config[model][filetype + "_sources"][
                                newdescr
                            ] = new_filename
                            if (
                                config[model][filetype + "_targets"][descr] == filename
                            ):  # source and target are identical if autocompleted
                                config[model][filetype + "_targets"][
                                    newdescr
                                ] = os.path.basename(new_filename)
                            else:
                                config[model][filetype + "_targets"][newdescr] = config[
                                    model
                                ][filetype + "_targets"][descr]
                            running_index += 1

                        del config[model][filetype + "_targets"][descr]
    return config


def target_subfolders(config):
    for filetype in config["general"]["all_model_filetypes"]:
        for model in config["general"]["valid_model_names"] + ["general"]:
            if filetype + "_targets" in config[model]:
                for descr, filename in config[model][filetype + "_targets"].items():
                    # * only in targets if denotes subfolder
                    # TODO: change with user_error()
                    if not descr in config[model][filetype + "_sources"]:
                        esm_parser.user_error(
                            "Filelists",
                            f"No source found for target ``{name}`` in model "
                            f"``{model}``\n"
                        )
                    if "*" in filename:
                        source_filename = os.path.basename(
                            config[model][filetype + "_sources"][descr]
                        )
                        # directory wildcards are given as /*, wildcards in filenames are handled
                        # seb-wahl: directory wildcards are given as /*, wildcards in filenames are handled
                        # in routine 'globbing' above, if we don't check here, wildcards are handled twice
                        # for files and hence filenames of e.g. restart files are screwed up.
                        if filename.endswith("/*"):
                            config[model][filetype + "_targets"][
                                descr
                            ] = filename.replace("*", source_filename)
                        elif "/" in filename:
                            # Return the correct target name
                            target_name = get_target_name_from_wildcard(
                                config, model, filename, filetype, descr
                            )
                            config[model][filetype + "_targets"][descr] = (
                                "/".join(filename.split("/")[:-1])
                                + "/"
                                + target_name
                            )
                        else:
                            # Return the correct target name
                            target_name = get_target_name_from_wildcard(
                                config, model, filename, filetype, descr
                            )
                            config[model][filetype + "_targets"][descr] = target_name
                    elif filename.endswith("/"):
                        source_filename = os.path.basename(
                            config[model][filetype + "_sources"][descr]
                        )
                        config[model][filetype + "_targets"][descr] = (
                            filename + source_filename
                        )

    return config


def get_target_name_from_wildcard(config, model, filename, filetype, descr):
    """
    Given a taget ``filename`` path containing a wildcard (``*``), and the ``descr``
    key to an specific file (i.e. the one that points to the actual specific file),
    returns the ``target_name`` basename for that specific file, taking care of
    changing the name from source to target, if both are compatible (same number of
    ``*``).

    For example::
      filename = "<new_name>.*.nc"
      model = "fesom"
      filetype = "restart_in"
      config["fesom"]["restart_in_sources_wild_card"] = "<old_name>.*.nc"
      config["fesom"]["restart_in_sources"]["<restart_file>_glob_1"] = "<old_name>.temp.nc"

    Then the outcome will be substitutting ``temp`` into the ``*`` of ``filename``::
      target_name = "<new_name>.temp.nc"

    Note:: MA: This is very poorly generalized and too specific to the current use.
    I volunteer myself to clean the ``filelists.py``, and make this method more
    general.

    Parameters
    ----------
    config : dict
        The experiment configuration
    model : str
        Target component/model
    filename : str
        Target path, with or without wildcards
    filetype : str
        Target file type
    descr : str
        Key pointing at the specific file, which target name needs to be constructed

    Returns
    -------
    targer_name : str
        Basename of the specific target file associated to the ``descr`` key

    Raises
    ------
    user_error : esm_parser.user_error
        If source and target wildcard patterns do not match (different number of ``*``
        in the patterns), raises a user friendly error
    """

    source_filename = os.path.basename(config[model][filetype + "_sources"][descr])
    target_filename = os.path.basename(filename)

    # Get the general description of the filename
    gen_descr = descr.split("_glob_")[0]
    # Load the wild cards from source and target and split them
    # at the *
    wild_card_source_all = config[model][
        f"{filetype}_sources_wild_card"
    ]
    wild_card_source = wild_card_source_all[gen_descr].split(
        "*"
    )
    wild_card_target = target_filename.split("*")
    # Check for syntax mistakes
    if len(wild_card_target) != len(wild_card_source):
        esm_parser.user_error(
            "Wild card",
            (
                "The wild card pattern of the source "
                + f"{wild_card_source} does not match with the "
                + f"target {wild_card_target}. Make sure the "
                + f"that the number of * are the same in both "
                + f"sources and targets."
            ),
        )
    # Loop through the pieces of the wild cards to create
    # the correct target name by substituting in the source
    # name
    target_name = source_filename
    for wcs, wct in zip(wild_card_source, wild_card_target):
        target_name = target_name.replace(wcs, wct)

    return target_name


def complete_restart_in(config):
    for model in config["general"]["valid_model_names"]:
        if (
            not config[model]["lresume"] and config["general"]["run_number"] == 1
        ):  # isn't that redundant? if run_number > 1 then lresume == True?
            if "restart_in_sources" in config[model]:
                del config[model]["restart_in_sources"]
            if "restart_in_targets" in config[model]:
                del config[model]["restart_in_targets"]
            if "restart_in_intermediate" in config[model]:
                del config[model]["restart_in_intermediate"]
        if "restart_in_sources" in config[model]:
            for category in list(config[model]["restart_in_sources"].keys()):
                if not config[model]["restart_in_sources"][category].startswith("/"):
                    config[model]["restart_in_sources"][category] = (
                        config[model]["parent_restart_dir"]
                        + config[model]["restart_in_sources"][category]
                    )
    return config


def assemble_intermediate_files_and_finalize_targets(config):
    for filetype in config["general"]["all_model_filetypes"]:
        for model in config["general"]["valid_model_names"] + ["general"]:
            if filetype + "_targets" in config[model]:
                if not filetype + "_intermediate" in config[model]:
                    config[model][filetype + "_intermediate"] = {}
                for category in config[model][filetype + "_targets"]:
                    target_name = config[model][filetype + "_targets"][category]

                    interm_dir = (
                        config[model]["thisrun_" + filetype + "_dir"] + "/"
                    ).replace("//", "/")
                    if filetype in config["general"]["out_filetypes"]:
                        target_dir = (
                            config[model]["experiment_" + filetype + "_dir"] + "/"
                        ).replace("//", "/")
                        source_dir = (
                            config["general"]["thisrun_work_dir"] + "/"
                        ).replace("//", "/")
                        if not config[model][filetype + "_sources"][
                            category
                        ].startswith("/"):
                            config[model][filetype + "_sources"][category] = (
                                source_dir
                                + config[model][filetype + "_sources"][category]
                            )
                    else:
                        target_dir = (config["general"]["thisrun_work_dir"]).replace(
                            "//", "/"
                        )

                    config[model][filetype + "_intermediate"][category] = (
                        interm_dir + target_name
                    )
                    config[model][filetype + "_targets"][category] = (
                        target_dir + target_name
                    )

    return config


def find_valid_year(config, year):
    for entry in config:
        min_val = -50000000000
        max_val = 500000000000

        from_info = float(config[entry].get("from", min_val))
        to_info = float(config[entry].get("to", max_val))

        if from_info <= year <= to_info:
            return entry

    error_type = "Year Error"
    error_text = f"Sorry, no entry found for year {year} in config {config}"
    esm_parser.user_error(error_type, error_text)


def replace_year_placeholder(config):
    for filetype in config["general"]["all_model_filetypes"]:
        for model in config["general"]["valid_model_names"] + ["general"]:
            if filetype + "_targets" in config[model]:
                if filetype + "_additional_information" in config[model]:
                    for file_category in config[model][
                        filetype + "_additional_information"
                    ]:
                        if file_category in config[model][filetype + "_targets"]:

                            all_years = [config["general"]["current_date"].year]

                            if (
                                "need_timestep_before"
                                in config[model][filetype + "_additional_information"][
                                    file_category
                                ]
                            ):
                                all_years.append(config["general"]["prev_date"].year)
                            if (
                                "need_timestep_after"
                                in config[model][filetype + "_additional_information"][
                                    file_category
                                ]
                            ):
                                all_years.append(config["general"]["next_date"].year)
                            if (
                                "need_year_before"
                                in config[model][filetype + "_additional_information"][
                                    file_category
                                ]
                            ):
                                all_years.append(
                                    config["general"]["current_date"].year - 1
                                )
                            if (
                                "need_year_after"
                                in config[model][filetype + "_additional_information"][
                                    file_category
                                ]
                            ):
                                all_years.append(
                                    config["general"]["current_date"].year + 1
                                )

                            if (
                                "need_2years_before"
                                in config[model][filetype + "_additional_information"][
                                    file_category
                                ]
                            ):
                                all_years.append(
                                    config["general"]["current_date"].year - 2
                                )

                            if (
                                "need_2years_after"
                                in config[model][filetype + "_additional_information"][
                                    file_category
                                ]
                            ):
                                all_years.append(
                                    config["general"]["current_date"].year + 2
                                )

                            all_years = list(
                                dict.fromkeys(all_years)
                            )  # removes duplicates

                            # loop over all years (including year_before & after)
                            # change replace the @YEAR@ variable with the
                            # corresponding year
                            for year in all_years:
                                new_category = file_category + "_year_" + str(year)

                                # if the source contains 'from' or 'to' information
                                # then they have a dict type
                                if (
                                    type(
                                        config[model][filetype + "_sources"][
                                            file_category
                                        ]
                                    )
                                    == dict
                                ):

                                    # process the 'from' and 'to' information in
                                    # file sources and targets
                                    config[model][filetype + "_sources"][
                                        new_category
                                    ] = find_valid_year(
                                        config[model][filetype + "_sources"][
                                            file_category
                                        ],
                                        year,
                                    )

                                    config[model][filetype + "_targets"][
                                        new_category
                                    ] = config[model][filetype + "_targets"][
                                        file_category
                                    ]

                                    # replace @YEAR@ in the targets
                                    if (
                                        "@YEAR@"
                                        in config[model][filetype + "_targets"][
                                            new_category
                                        ]
                                    ):
                                        new_target_name = config[model][
                                            filetype + "_targets"
                                        ][new_category].replace("@YEAR@", str(year))

                                        config[model][filetype + "_targets"][
                                            new_category
                                        ] = new_target_name

                                    # replace @YEAR@ in the sources
                                    if (
                                        "@YEAR@"
                                        in config[model][filetype + "_sources"][
                                            new_category
                                        ]
                                    ):
                                        new_source_name = config[model][
                                            filetype + "_sources"
                                        ][new_category].replace("@YEAR@", str(year))

                                        config[model][filetype + "_sources"][
                                            new_category
                                        ] = new_source_name

                                # value is not a dictionary. Ie. it does not
                                # have `from` or `to` attributes. This else
                                # block preserves these sections in the config.
                                else:
                                    # create `new_category` from `file_category`
                                    config[model][filetype + "_sources"][
                                        new_category
                                    ] = config[model][filetype + "_sources"][
                                        file_category
                                    ]

                                    config[model][filetype + "_targets"][
                                        new_category
                                    ] = config[model][filetype + "_targets"][
                                        file_category
                                    ]

                                    # replace @YEAR@ in the targets
                                    if (
                                        "@YEAR@"
                                        in config[model][filetype + "_targets"][
                                            new_category
                                        ]
                                    ):
                                        new_target_name = config[model][
                                            filetype + "_targets"
                                        ][new_category].replace("@YEAR@", str(year))

                                        config[model][filetype + "_targets"][
                                            new_category
                                        ] = new_target_name

                                    # replace @YEAR@ in the sources
                                    if (
                                        "@YEAR@"
                                        in config[model][filetype + "_sources"][
                                            new_category
                                        ]
                                    ):
                                        new_source_name = config[model][
                                            filetype + "_sources"
                                        ][new_category].replace("@YEAR@", str(year))

                                        config[model][filetype + "_sources"][
                                            new_category
                                        ] = new_source_name

                                # end if
                            # end of the for year loop

                            # deniz: new additions for @YEAR_1850@
                            # these are the Kinne aerosol files for the background
                            # aerosol concentration. They are needed for years
                            # 1849, 1850, and 1851. All these 3 files are the same
                            # and ECHAM needs them
                            if (
                                "@YEAR_1850@"
                                in config[model][filetype + "_targets"][file_category]
                            ):
                                # only target name is changed since source file is for a fixed year (1850)
                                for year in [1849, 1850, 1851]:
                                    new_category = file_category + "_year_" + str(year)

                                    # add the sources and targets to the config
                                    config[model][filetype + "_sources"][
                                        new_category
                                    ] = config[model][filetype + "_sources"][
                                        file_category
                                    ]

                                    config[model][filetype + "_targets"][
                                        new_category
                                    ] = config[model][filetype + "_targets"][
                                        file_category
                                    ]

                                    # construct the file target and add this to the config
                                    new_target_name = config[model][
                                        filetype + "_targets"
                                    ][new_category].replace("@YEAR_1850@", str(year))

                                    config[model][filetype + "_targets"][
                                        new_category
                                    ] = new_target_name

                            del config[model][filetype + "_sources"][file_category]
                            del config[model][filetype + "_targets"][file_category]
                # end of if additonal information

                year = config["general"]["current_date"].year

                for file_category in config[model][filetype + "_targets"]:

                    if (
                        type(config[model][filetype + "_sources"][file_category])
                        == dict
                    ):
                        config[model][filetype + "_sources"][
                            file_category
                        ] = find_valid_year(
                            config[model][filetype + "_sources"][file_category], year
                        )
                    if "@YEAR@" in config[model][filetype + "_targets"][file_category]:
                        new_target_name = config[model][filetype + "_targets"][
                            file_category
                        ].replace("@YEAR@", str(year))

                        config[model][filetype + "_targets"][
                            file_category
                        ] = new_target_name

                    if "@YEAR@" in config[model][filetype + "_sources"][file_category]:
                        new_source_name = config[model][filetype + "_sources"][
                            file_category
                        ].replace("@YEAR@", str(year))

                        config[model][filetype + "_sources"][
                            file_category
                        ] = new_source_name

            # end of if filetype in target
        # end of model loop
    # end of filetype loop
    return config


def log_used_files(config):
    if config["general"]["verbose"]:
        print("\n::: Logging used files", flush=True)
    filetypes = config["general"]["relevant_filetypes"]
    expid = config["general"]["expid"]
    it_coupled_model_name = config["general"]["iterative_coupled_model"]
    datestamp = config["general"]["run_datestamp"]
    for model in config["general"]["valid_model_names"] + ["general"]:
        thisrun_config_dir = config[model]["thisrun_config_dir"]
        # this file contains the files used in the experiment
        flist_file = (
            f"{thisrun_config_dir}/"
            f"{expid}_{it_coupled_model_name}filelist_{datestamp}"
        )

        with open(flist_file, "w") as flist:
            flist.write(
                f"These files are used for \n"
                f"experiment {config['general']['expid']}\n"
                f"component {model}\n"
                f"date {config['general']['run_datestamp']}"
            )
            flist.write("\n")
            flist.write(80 * "-")
            for filetype in filetypes:
                if filetype + "_sources" in config[model]:
                    flist.write("\n" + filetype.upper() + ":\n")
                    for category in config[model][filetype + "_sources"]:
                        #                        esm_parser.pprint_config(config[model])
                        flist.write(
                            "\nSource: "
                            + config[model][filetype + "_sources"][category]
                        )
                        flist.write(
                            "\nExp Tree: "
                            + config[model][filetype + "_intermediate"][category]
                        )
                        flist.write(
                            "\nTarget: "
                            + config[model][filetype + "_targets"][category]
                        )
                        if config["general"]["verbose"]:
                            print(flush=True)
                            print(f"::: logging file category: {category}")
                            print(
                                (
                                    f"- source: "
                                    f'{config[model][filetype + "_sources"][category]}'
                                ),
                                flush=True,
                            )
                            print(
                                (
                                    f"- target: "
                                    f'{config[model][filetype + "_targets"][category]}'
                                ),
                                flush=True,
                            )
                            helpers.print_datetime(config)
                        flist.write("\n")
                flist.write("\n")
                flist.write(80 * "-")
    return config


def check_for_unknown_files(config):
    # files = os.listdir(config["general"]["thisrun_work_dir"])
    all_files = glob.iglob(
        config["general"]["thisrun_work_dir"] + "**/*", recursive=True
    )

    known_files = [
        config["general"]["thisrun_work_dir"] + "/" + "hostfile_srun",
        config["general"]["thisrun_work_dir"] + "/" + "namcouple",
        config["general"]["thisrun_work_dir"] + "/" + "coupling.xml",
    ]

    for filetype in config["general"]["all_model_filetypes"]:
        for model in config["general"]["valid_model_names"] + ["general"]:
            if filetype + "_sources" in config[model]:
                known_files += list(config[model][filetype + "_sources"].values())
                known_files += list(config[model][filetype + "_targets"].values())

    known_files = [os.path.realpath(known_file) for known_file in known_files]
    known_files = list(dict.fromkeys(known_files))

    if not "unknown_sources" in config["general"]:
        config["general"]["unknown_sources"] = {}
        config["general"]["unknown_targets"] = {}
        config["general"]["unknown_intermediate"] = {}

    unknown_files = []
    index = 0

    for thisfile in all_files:

        if os.path.realpath(thisfile) in known_files + unknown_files:
            continue
        config["general"]["unknown_sources"][index] = os.path.realpath(thisfile)
        config["general"]["unknown_targets"][index] = os.path.realpath(
            thisfile
        ).replace(
            os.path.realpath(config["general"]["thisrun_work_dir"]),
            os.path.realpath(config["general"]["experiment_unknown_dir"]),
        )
        config["general"]["unknown_intermediate"][index] = os.path.realpath(
            thisfile
        ).replace(
            os.path.realpath(config["general"]["thisrun_work_dir"]),
            os.path.realpath(config["general"]["thisrun_unknown_dir"]),
        )

        unknown_files.append(os.path.realpath(thisfile))

        index += 1
        print("Unknown file in work: " + os.path.realpath(thisfile), flush=True)
        helpers.print_datetime(config)

    return config


def resolve_symlinks(file_source):
    if os.path.islink(file_source):
        points_to = os.path.realpath(file_source)

        # deniz: check if file links to itself. In UNIX
        # ln -s endless_link endless_link is a valid command
        if os.path.abspath(file_source) == points_to:
            if config["general"]["verbose"]:
                print(f"file {file_source} links to itself", flush=True)
                helpers.print_datetime(config)
            return file_source

        # recursively find the file that the link is pointing to
        return resolve_symlinks(points_to)
    else:
        return file_source


def copy_files(config, filetypes, source, target):
    if config["general"]["verbose"]:
        print("\n::: Copying files", flush=True)
        helpers.print_datetime(config)

    successful_files = []
    missing_files = {}

    if source == "init":
        text_source = "sources"
    elif source == "thisrun":
        text_source = "intermediate"
    elif source == "work":
        text_source = "sources"

    if target == "thisrun":
        text_target = "intermediate"
    elif target == "work":
        text_target = "targets"

    for filetype in [filetype for filetype in filetypes if not filetype == "ignore"]:
        for model in config["general"]["valid_model_names"] + ["general"]:
            if filetype + "_" + text_source in config[model]:
                sourceblock = config[model][filetype + "_" + text_source]
                targetblock = config[model][filetype + "_" + text_target]
                for category in sourceblock:
                    movement_method = get_method(
                        get_movement(config, model, category, filetype, source, target)
                    )
                    file_source = os.path.normpath(sourceblock[category])
                    file_target = os.path.normpath(targetblock[category])
                    if config["general"]["verbose"]:
                        print(flush=True)
                        print(f"::: copying file category: {category}")
                        print(f"- source: {file_source}", flush=True)
                        print(f"- target: {file_target}", flush=True)
                        helpers.print_datetime(config)
                    if file_source == file_target:
                        if config["general"]["verbose"]:
                            print(
                                f"Source and target paths are identical, skipping {file_source}",
                                flush=True,
                            )
                            helpers.print_datetime(config)
                        continue
                    dest_dir = os.path.dirname(file_target)
                    file_source = resolve_symlinks(file_source)
                    if not os.path.isdir(file_source):
                        try:
                            if not os.path.isdir(dest_dir):
                                # MA: ``os.makedirs`` creates the specified directory
                                # and the parent directories if the last don't exist
                                # (same as with ``mkdir -p <directory>>``)
                                os.makedirs(dest_dir)
                            if not os.path.isfile(file_source):
                                print(
                                    f"WARNING: File not found: {file_source}",
                                    flush=True,
                                )
                                helpers.print_datetime(config)
                                missing_files.update({file_target: file_source})
                                continue
                            if os.path.isfile(file_target) and filecmp.cmp(
                                file_source, file_target
                            ):
                                if config["general"]["verbose"]:
                                    print(
                                        f"Source and target file are identical, skipping {file_source}",
                                        flush=True,
                                    )
                                    helpers.print_datetime(config)
                                continue
                            movement_method(file_source, file_target)
                            # shutil.copy2(file_source, file_target)
                            successful_files.append(file_source)
                        except IOError:
                            print(
                                f"Could not copy {file_source} to {file_target} for unknown reasons.",
                                flush=True,
                            )
                            helpers.print_datetime(config)
                            missing_files.update({file_target: file_source})

    if missing_files:
        if not "files_missing_when_preparing_run" in config["general"]:
            config["general"]["files_missing_when_preparing_run"] = {}
        if config["general"]["verbose"]:
            print("\n\nWARNING: These files were missing:")
            for missing_file in missing_files:
                print(f"- missing source: {missing_files[missing_file]}", flush=True)
                print(f"- missing target: {missing_file}", flush=True)
                helpers.print_datetime(config)
        config["general"]["files_missing_when_preparing_run"].update(missing_files)
    return config


def filter_allowed_missing_files(config):
    """
    Filters the general.files_missing_when_preparing_run dictionary to move any
    allowed missing files to a seperate dictionary.


    This function can be included to mark specific files as "allowed to be
    missing". A list of allowed missing files should be put into the model
    configuration::

        echam:
            allowed_missing_files:
                - "*restart*jsbid*"
                - "unit.*"

    The allowed missing files may either be globs or regular expressions. The
    filename (without full path) on either the source or the target is used for
    matching. Therefore, if a file is renamed in the process of being moved
    from the pool filesystem to the experiment tree, or to the work folder, an
    exclusion can be made both with the source name or with the target name.

    Parameters
    ----------
    config : dict
        The experiment configuration

    Returns
    -------
    config : dict
    """
    allowed_missing_files = config["general"].setdefault("allowed_missing_files", {})
    missing_files = config["general"].get("files_missing_when_preparing_run", {})
    # TODO(PG): Replace with logger statements
    if config["general"].get("verbose", False):
        print("Currently known missing files:")
        for k, v in missing_files.items():
            print(f"source: {k} --> target: {v}")
    remove_missing_files = []
    for missing_file_source, missing_file_target in missing_files.items():
        missing_file_source_fname = pathlib.Path(missing_file_source).name
        missing_file_target_fname = pathlib.Path(missing_file_target).name
        for model in config["general"]["valid_model_names"] + ["general"]:
            for allowed_missing_pattern in config[model].get(
                "allowed_missing_files", []
            ):
                if (
                    re.match(allowed_missing_pattern, missing_file_source_fname)
                    or missing_file_source_fname in glob.glob(allowed_missing_pattern)
                    or re.match(allowed_missing_pattern, missing_file_target_fname)
                    or missing_file_target_fname in glob.glob(allowed_missing_pattern)
                ):
                    # TODO(PG): Replace with logger statements
                    if config["general"].get("verbose", False):
                        print(
                            f"Detected allowed missing file with {allowed_missing_pattern}"
                        )
                        print("Adding to allowed missing files:")
                        print(f"source: {missing_file_source}")
                        print(f"target: {missing_file_target}")
                    remove_missing_files.append(missing_file_source)
                    allowed_missing_files.update(
                        {missing_file_source: missing_file_target}
                    )
    for remove_missing in remove_missing_files:
        missing_files.pop(remove_missing)
    return config


def report_missing_files(config):
    # this list is populated by the ``copy_files`` function in filelists.py
    config = _check_fesom_missing_files(config)
    if "files_missing_when_preparing_run" in config["general"]:
        if not config["general"]["files_missing_when_preparing_run"] == {}:
            print("MISSING FILES:", flush=True)
        for missing_file in config["general"]["files_missing_when_preparing_run"]:
            print(flush=True)
            print(
                f'- missing source: {config["general"]["files_missing_when_preparing_run"][missing_file]}',
                flush=True,
            )
            print(f"- missing target: {missing_file}", flush=True)
            helpers.print_datetime(config)
        if not config["general"]["files_missing_when_preparing_run"] == {}:
            print(80 * "=")
        print()
    return config


def _check_fesom_missing_files(config):
    """
    Checks for missing files in FESOM namelist.config

    Parameters
    ----------
    config : dict
        The experiment configuration

    Returns
    -------
    config : dict
    """
    if "fesom" in config["general"]["valid_model_names"]:
        namelist_config = f90nml.read(
            os.path.join(config["general"]["thisrun_work_dir"], "namelist.config")
        )
        for path_key, path in namelist_config["paths"].items():
            if path:  # Remove empty strings
                if not os.path.exists(path):
                    if "files_missing_when_preparing_run" not in config["general"]:
                        config["general"]["files_missing_when_preparing_run"] = {}
                    config["general"]["files_missing_when_preparing_run"][
                        path_key + " (from namelist.config in FESOM)"
                    ] = path
    return config


# FILE MOVEMENT METHOD STUFF


def create_missing_file_movement_entries(config):
    for model in config["general"]["valid_model_names"] + ["general"]:
        if not "file_movements" in config[model]:
            config[model]["file_movements"] = {}
        for filetype in config["general"]["all_model_filetypes"] + [
            "scripts",
            "unknown",
        ]:
            if not filetype in config[model]["file_movements"]:
                config[model]["file_movements"][filetype] = {}
    return config


def complete_one_file_movement(config, model, filetype, movement, movetype):
    if not movement in config[model]["file_movements"][filetype]:
        config[model]["file_movements"][filetype][movement] = movetype
    return config


def get_method(movement):
    if movement == "copy":
        return shutil.copy2
    elif movement == "link":
        return os.symlink
    elif movement == "move":
        return os.rename
    print("Unknown file movement type, using copy (safest option).", flush=True)
    return shutil.copy2


def complete_all_file_movements(config):

    mconfig = config["general"]
    general_file_movements = copy.deepcopy(mconfig.get("file_movements", {}))
    if "defaults.yaml" in mconfig:
        if "per_model_defaults" in mconfig["defaults.yaml"]:
            if "file_movements" in mconfig["defaults.yaml"]["per_model_defaults"]:
                mconfig["file_movements"] = copy.deepcopy(
                    mconfig["defaults.yaml"]["per_model_defaults"]["file_movements"]
                )
                del mconfig["defaults.yaml"]["per_model_defaults"]["file_movements"]
    if not "file_movements" in mconfig:
        mconfig["file_movements"] = {}
    # General ``file_movements`` win over default ones
    esm_parser.dict_merge(mconfig["file_movements"], general_file_movements)

    config = create_missing_file_movement_entries(config)

    for model in config["general"]["valid_model_names"] + ["general"]:
        mconfig = config[model]
        if "file_movements" in mconfig:
            for filetype in config["general"]["all_model_filetypes"] + [
                "scripts",
                "unknown",
            ]:
                if filetype in mconfig["file_movements"]:
                    if "all_directions" in mconfig["file_movements"][filetype]:
                        movement_type = mconfig["file_movements"][filetype][
                            "all_directions"
                        ]
                        for movement in [
                            "init_to_exp",
                            "exp_to_run",
                            "run_to_work",
                            "work_to_run",
                        ]:
                            config = complete_one_file_movement(
                                config, model, filetype, movement, movement_type
                            )
                        del mconfig["file_movements"][filetype]["all_directions"]

    for model in config["general"]["valid_model_names"] + ["general"]:
        mconfig = config[model]
        if "file_movements" in mconfig:
            # Complete movements with general
            esm_parser.new_deep_update(
                mconfig["file_movements"], config["general"].get("file_movements", {})
            )

            # Complete file specific movements with ``all_directions``
            for file_in_fm in mconfig["file_movements"]:
                # If it is a specific file, and not a file type
                if file_in_fm not in (
                    config["general"]["all_model_filetypes"] + ["scripts", "unknown"]
                ):
                    # Check syntax for restart files
                    if file_in_fm in mconfig.get(
                        "restart_in_files", {}
                    ) or file_in_fm in mconfig.get("restart_out_files", {}):
                        esm_parser.user_error(
                            "Movement direction not specified",
                            f"'{model}.file_movements.{file_in_fm}' refers to a "
                            + "restart file which can be moved/copied/link in two "
                            + "directions, into the 'work' folder and out of the "
                            + "'work' folder. Please, add the direction '_in' or "
                            + f"'_out' to '{file_in_fm}':\n\n{model}:\n    "
                            + f"file_movements:\n        {file_in_fm}_<in/out>:\n"
                            + f"            [ ... ]",
                        )
                    # Solve ``all_directions``
                    file_spec_movements = mconfig["file_movements"][file_in_fm]
                    if "all_directions" in file_spec_movements:
                        movement_type = file_spec_movements["all_directions"]
                        for movement in [
                            "init_to_exp",
                            "exp_to_run",
                            "run_to_work",
                            "work_to_run",
                        ]:
                            config = complete_one_file_movement(
                                config, model, file_in_fm, movement, movement_type
                            )
                        del mconfig["file_movements"][file_in_fm]["all_directions"]

            if "default" in mconfig["file_movements"]:
                if "all_directions" in mconfig["file_movements"]["default"]:
                    movement_type = mconfig["file_movements"]["default"][
                        "all_directions"
                    ]
                    for movement in [
                        "init_to_exp",
                        "exp_to_run",
                        "run_to_work",
                        "work_to_run",
                    ]:
                        config = complete_one_file_movement(
                            config, model, "default", movement, movement_type
                        )
                    del mconfig["file_movements"]["default"]["all_directions"]

                for movement in mconfig["file_movements"]["default"]:
                    movement_type = mconfig["file_movements"]["default"][movement]
                    for filetype in config["general"]["all_model_filetypes"] + [
                        "scripts",
                        "unknown",
                    ]:
                        config = complete_one_file_movement(
                            config, model, filetype, movement, movement_type
                        )
                del mconfig["file_movements"]["default"]
    return config


def get_movement(config, model, category, filetype, source, target):
    # Remove globing strings from category
    if isinstance(category, str):
        category = category.split("_glob_")[0]
    # Two type of directions are needed for restarts, therefore, the categories need an
    # "_in" or "_out" at the end.
    if filetype == "restart_in":
        category = f"{category}_in"
    elif filetype == "restart_out":
        category = f"{category}_out"
    # File specific movements
    file_spec_movements = config[model]["file_movements"].get(category, {})
    # Movements associated to ``filetypes``
    file_type_movements = config[model]["file_movements"][filetype]
    if source == "init":
        # Get the model-specific reusable_filetypes, if not existing, get the
        # general ones
        model_reusable_filetypes = config[model].get(
            "reusable_filetypes", config["general"]["reusable_filetypes"]
        )

        if (
            config["general"]["run_number"] == 1
            or filetype not in model_reusable_filetypes
        ):
            return file_spec_movements.get(
                "init_to_exp", file_type_movements["init_to_exp"]
            )
        else:
            return file_spec_movements.get(
                "exp_to_run", file_type_movements["exp_to_run"]
            )
    elif source == "work":
        return file_spec_movements.get(
            "work_to_run", file_type_movements["work_to_run"]
        )
    elif source == "thisrun" and target == "work":
        return file_spec_movements.get(
            "run_to_work", file_type_movements["run_to_work"]
        )
    else:
        # This should NOT happen
        print(f"Error: Unknown file movement from {source} to {target}", flush=True)
        helpers.print_datetime(config)
        sys.exit(42)


def assemble(config):
    config = complete_all_file_movements(config)
    config = rename_sources_to_targets(config)
    config = choose_needed_files(config)
    config = complete_targets(config)
    config = complete_sources(config)
    config = reuse_sources(config)
    config = replace_year_placeholder(config)

    config = complete_restart_in(config)
    config = globbing(config)
    config = target_subfolders(config)
    config = assemble_intermediate_files_and_finalize_targets(config)
    return config



import signal
import sys
import os


def signal_listener():
    """Listens to the signals (eg. CTRL-C press, ...) and calls the handler
    for that signal"""

    # 1) handler for the SIGINT (CTRL-C)
    signal.signal(signal.SIGINT, handle_sigint)

    # other signal handlers may go here, eg. SIGHUP, SIGSTOP, ...)
    # each signal is handled by a function handle_<signal>, eg. handle_sigint()


def handle_sigint(signum, stack_frame):
    """Handles the SIGINT signal (aka CTRL-C press on UNIX/LINUX)

    Parameters
    ----------
    signum : int
        signal number.
    stack_frame : Python frame object
        current stack frame

    Notes
    -----
        More info: https://docs.python.org/3/library/signal.html
    """
    print()  # print a new line after ^C character (CTRL-C)
    sys.stderr.write("ctrl c is pressed. Terminating\n")

    # exit with 2 according to https://en.wikipedia.org/wiki/Signal_(IPC)
    sys.exit(signal.SIGINT.value)


if __name__ == "__main__":
    pass
